from __future__ import annotations

import importlib.metadata
import importlib.util
import logging
import os
import site
import subprocess
import sys
from pathlib import Path
from typing import Any

from hpoglue import Problem
from packaging import version

logger = logging.getLogger(__name__)


HPOSUITR_REPO = "github.com/automl/hposuite.git"
HPOSUITE_EDITABLE = Path(__file__).parent.parent
HPOSUITE_PYPI = "hposuite"
HPOSUITE_GIT_SSH_INSTALL = "git+ssh://git@github.com/automl/hposuite.git"

class GlueWrapperFunctions:
    """A collection of wrapper functions around certain hpoglue methods."""

    @staticmethod
    def problem_from_dict(data: dict[str, Any]) -> Problem:
        """Convert a dictionary to a Problem instance."""
        from hposuite.benchmarks import BENCHMARKS
        from hposuite.optimizers import OPTIMIZERS

        return Problem.from_dict(
            data=data,
            benchmarks_dict=BENCHMARKS,
            optimizers_dict=OPTIMIZERS,
        )


class HiddenPrints:  # noqa: D101
    def __enter__(self):
        self._original_stdout = sys.stdout
        from pathlib import Path
        sys.stdout = Path(os.devnull).open("w")

    def __exit__(self, exc_type, exc_val, exc_tb):
        sys.stdout.close()
        sys.stdout = self._original_stdout



def is_package_installed(package_name: str) -> bool:
    """Check if a package is installed in the current virtual environment."""
    # Extract version constraint if present
    version_constraints = ["==", ">=", "<=", ">", "<"]
    version_check = None
    for constraint in version_constraints:
        if constraint in package_name:
            package_name, version_spec = package_name.split(constraint)
            version_check = (constraint, version_spec)
            break

    # Normalize package name (replace hyphens with underscores)
    package_name = package_name.replace("-", "_")

    # Remove dependencies from the package name
    package_name = package_name.split("[")[0]

    # Get the site-packages directory of the current virtual environment
    venv_site_packages = site.getsitepackages() if hasattr(site, "getsitepackages") else []
    venv_prefix = sys.prefix  # Virtual environment root

    # Check if the package is installed in the virtual environment
    for site_package_path in venv_site_packages:
        package_path = Path(site_package_path) / package_name

        # Check if the package exists in the site-packages directory
        if package_path.exists() and venv_prefix in str(package_path):
            installed_version = importlib.metadata.version(package_name)
            if version_check:
                constraint, required_version = version_check
                return _check_package_version(installed_version, required_version, constraint)
            return True

        # Check if package is installed as different name (e.g., .dist-info or .egg-info)
        dist_info_pattern = f"{package_name}*"
        dist_info_paths = list(Path(site_package_path).glob(dist_info_pattern))
        if dist_info_paths:
            dist_info_name = dist_info_paths[0].name.replace(".dist-info", "") \
                .replace(".egg-info", "")
            installed_version = dist_info_name.split("-")[-1]
            if version_check:
                constraint, required_version = version_check
                return _check_package_version(installed_version, required_version, constraint)
            return True

    return False

def _check_package_version(
    installed_version: str,
    required_version: str,
    check_key: str,
):
    """Check if the installed package version satisfies the required version."""
    installed_version = version.parse(installed_version)
    required_version = version.parse(required_version)
    match check_key:
        case "==":
            return installed_version == required_version
        case ">=":
            return installed_version >= required_version
        case "<=":
            return installed_version <= required_version
        case ">":
            return installed_version > required_version
        case "<":
            return installed_version < required_version


def get_current_installed_hposuite_version() -> str:
    """Retrieve the currently installed version of hposuite."""
    cmd = ["pip", "show", "hposuite"]
    logger.debug(cmd)
    output = subprocess.run(  # noqa: S603
        cmd,
        check=True,
        capture_output=True,
        text=True,
    )
    lines = output.stdout.strip().splitlines()
    for line in lines:
        if "Version: " in line:
            return line.split(": ")[1]

    raise RuntimeError(f"Could not find hposuite version in {lines}.")


def get_current_ConfigSpace_version() -> version.Version:  # noqa: N802
    """Retrieve the currently installed version of ConfigSpace."""
    pkg_dict = {dist.name: dist.version for dist in importlib.metadata.distributions()}
    if "ConfigSpace" in pkg_dict:
        return version.parse(pkg_dict["ConfigSpace"])
    raise RuntimeError("ConfigSpace is not installed.")


def compare_installed_CS_version_vs_required(required_version: str) -> str:  # noqa: N802
    """Compare the installed ConfigSpace version with the required version."""
    installed_version = get_current_ConfigSpace_version()
    required_version = version.parse(required_version)
    if installed_version < required_version:
        return "<"
    if installed_version == required_version:
        return "=="
    return ">"


from __future__ import annotations

import hashlib
import logging
import warnings
from collections.abc import Iterable, Mapping
from dataclasses import dataclass, field
from itertools import product
from pathlib import Path
from typing import TYPE_CHECKING, Any, Literal, TypeAlias

import numpy as np
import yaml
from hpoglue import BenchmarkDescription, FunctionalBenchmark, Optimizer, Problem
from hpoglue.utils import configpriors_to_dict, dict_to_configpriors

from hposuite.benchmarks import BENCHMARKS
from hposuite.constants import DEFAULT_STUDY_DIR
from hposuite.optimizers import OPTIMIZERS
from hposuite.run import Run
from hposuite.utils import HPOSUITE_EDITABLE

if TYPE_CHECKING:
    from hpoglue.budget import BudgetType

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.DEBUG)

smac_logger = logging.getLogger("smac")
smac_logger.setLevel(logging.ERROR)



OptWithHps: TypeAlias = tuple[type[Optimizer], Mapping[str, Any]]

BenchWith_Objs_Fids: TypeAlias = tuple[BenchmarkDescription, Mapping[str, Any]]

GLOBAL_SEED = 42


@dataclass(kw_only=True)
class Study:
    """A class to represent a hposuite Study consisting of multiple Runs."""

    name: str | None = None
    """The name of the study."""

    output_dir: str | Path = DEFAULT_STUDY_DIR
    """The outer directory where all Studies are stored"""

    study_dir: Path = field(init=False)
    """The directory for storing the Study results.
        study_dir = output_dir/study_name
    """

    study_yaml_path: Path = field(init=False)
    """The path to the study configuration YAML file."""

    optimizers: list[OptWithHps] = field(init=False)
    """The list of optimizers used in the study."""

    benchmarks: list[BenchWith_Objs_Fids] = field(init=False)
    """The benchmarks used in the study."""

    experiments: list[Run]
    """The list of experiments in the study."""

    seeds: Iterable[int] | int | None = None
    """The seeds used in the study."""

    num_seeds: int = 1
    """The number of seeds used in the study."""

    budget: int = 50
    """The budget for the study."""

    group_by: Literal["opt", "bench", "opt_bench", "seed", "mem"] | None = None
    """The grouping method to dump run commands for the study."""

    continuations: bool = field(init=False)
    """Whether to use continuations for the study."""

    def __post_init__(self):  # noqa: C901

        self.optimizers = []
        self.benchmarks = []
        seeds = set()

        opt_keys = []
        bench_keys = {}
        continuations = 0
        for run in self.experiments:
            continuations += run.problem.continuations
            opt_name = run.name.split("benchmark")[0]
            if opt_name not in opt_keys:
                self.optimizers.append(
                    (
                        run.optimizer,
                        run.optimizer_hyperparameters or {},
                    )
                )
                opt_keys.append(opt_name)

            # TODO: This doesn't uniquely identify runs with priors (Problem level issue).

            _store_val = {
                "objectives": run.problem.get_objectives(),
                "fidelities": run.problem.get_fidelities(),
                "costs": run.problem.get_costs(),
                "priors": run.problem.priors,
            }

            bench = bench_keys.setdefault(run.benchmark.name, (run.benchmark, []))

            if _store_val not in bench[1]:
                for variant in bench[1]:
                    if variant["objectives"] == _store_val["objectives"]:
                        for key in ["fidelities", "costs", "priors"]:
                            variant[key] = variant[key] or _store_val[key]
                        break
                else:
                    bench[1].append(_store_val)

            seeds.add(run.seed)

        for _, v in bench_keys.items():
            [self.benchmarks.append(
                (
                    v[0],
                    var,
                )
            ) for var in v[1]]

        self.continuations = continuations > 0

        if self.seeds is None:
            self.seeds = list(seeds)
        self.num_seeds = len(self.seeds) if isinstance(self.seeds, Iterable) else 1

        name_parts: list[str] = []
        name_parts.append(";".join([f"{opt[0].name}{opt[-1]}" for opt in self.optimizers]))
        name_parts.append(";".join([f"{bench[0].name}{bench[-1]}" for bench in self.benchmarks]))
        name_parts.append(f"seeds={self.seeds}")
        name_parts.append(f"budget={self.budget}")

        if self.name is None:
            self.name = hashlib.sha256((".".join(name_parts)).encode()).hexdigest()

        self.study_dir = self.output_dir / self.name
        self.study_yaml_path = self.study_dir / "study_config.yaml"
        self.write_yaml()
        logger.info(f"Created study at {self.study_dir.absolute()}")


        if len(self.experiments) > 1:
            self._dump_runs(
                group_by=self.group_by,
                exp_dir=self.study_dir,
            )


    def _update_study(
        self,
        *,
        new_seeds: Iterable[int],
    ) -> None:
        """Update the study with new seeds."""
        more_experiments = Study.generate(
            optimizers=self.optimizers,
            benchmarks=self.benchmarks,
            budget=self.budget,
            seeds=new_seeds,
            continuations=self.continuations,
        )

        self.seeds.extend(new_seeds)
        self.num_seeds += len(new_seeds)
        self.experiments.extend(more_experiments)
        self.write_yaml()



    def to_dict(self) -> dict[str, Any]:
        """Convert the study to a dictionary."""
        for run in self.experiments:
            run._set_paths(self.study_dir)
            run.write_yaml()

        _optimizers = [{"name": opt[0].name, "hyperparameters": opt[1]} for opt in self.optimizers]
        _benchmarks = []
        for bench in self.benchmarks:
            _priors = None
            if bench[1]["priors"]:
                _priors = configpriors_to_dict(bench[1]["priors"])
                _priors = list(_priors)
            _benchmarks.append(
                {
                    "name": bench[0].name,
                    "objectives": bench[1]["objectives"],
                    "fidelities": bench[1]["fidelities"],
                    "costs": bench[1]["costs"],
                    "priors": _priors,
                }
            )

        return {
            "study_name": self.name,
            "output_dir": str(self.output_dir.absolute()),
            "optimizers": _optimizers,
            "benchmarks": _benchmarks,
            "seeds": self.seeds,
            "num_seeds": self.num_seeds,
            "budget": self.budget,
            "continuations": self.continuations,
        }


    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> Study:  # noqa: C901, PLR0912
        """Create a Study object from a dictionary."""
        # TODO: Add documentation for usage
        # TODO: Test for all allowed types in Study.create_study()
        _optimizers = []
        for opt in data["optimizers"]:
            match opt:
                case Mapping():
                    assert "name" in opt, f"Optimizer name not found in {opt}"
                    if len(opt) == 1 or not opt.get("hyperparameters"):
                        _optimizers.append((opt["name"], {}))
                    else:
                        _optimizers.append(tuple(opt.values()))
                case str():
                    _optimizers.append((opt, {}))
                case tuple():
                    assert len(opt) <= 2, "Each Optimizer must only have a name and hyperparameters"  # noqa: PLR2004
                    assert isinstance(opt[0], str), "Expected str for optimizer name"
                    if len(opt) == 1:
                        _optimizers.append((opt[0], {}))
                    else:
                        assert isinstance(opt[1], Mapping), (
                            "Expected Mapping for Optimizer hyperparameters"
                        )
                        _optimizers.append(opt)
                case _:
                    raise ValueError(
                        f"Invalid type for optimizer: {type(opt)}. "
                        "Expected Mapping, str or tuple"
                    )

        _benchmarks = []
        for bench in data["benchmarks"]:
            match bench:
                case Mapping():
                    assert "name" in bench, f"Benchmark name not found in {bench}"
                    assert "objectives" in bench, f"Benchmark objectives not found in {bench}"
                    _benchmarks.append(
                        (
                            bench["name"],
                            {
                                "objectives": bench["objectives"],
                                "fidelities": bench.get("fidelities"),
                                "costs": bench.get("costs"),
                                "priors": bench.get("priors"),
                            }
                        )
                    )
                case str():
                    _benchmarks.append(bench)
                case tuple():
                    assert len(bench) == 2, "Each Benchmark must only have a name and a Mapping"  # noqa: PLR2004
                    assert isinstance(bench[0], str), "Expected str for benchmark name"
                    assert isinstance(bench[1], Mapping), (
                        "Expected Mapping for Benchmark objectives and fidelities"
                    )
                    _benchmarks.append(bench)
                case _:
                    raise ValueError(
                        f"Invalid type for benchmark: {type(bench)}. "
                        "Expected Mapping, str or tuple"
                    )


        return create_study(
            name=data.get("study_name"),
            output_dir=data.get("output_dir"),
            optimizers=_optimizers,
            benchmarks=_benchmarks,
            seeds=data.get("seeds"),
            num_seeds=data.get("num_seeds", 1),
            budget=data.get("budget", 50),
            group_by=data.get("group_by"),
            on_error=data.get("on_error", "warn"),
        )


    @classmethod
    def from_yaml(cls, yaml_file: Path | str) -> Study:
        """Create a Study instance from a YAML file.

        Args:
            yaml_file: The path to the YAML file containing the study data.

        Returns:
            An instance of the Study class populated with data from the YAML file.
        """
        if isinstance(yaml_file, str):
            yaml_file = Path(yaml_file)
        with yaml_file.open("r") as file:
            data = yaml.safe_load(file)
        return cls.from_dict(data)


    def write_yaml(self) -> None:
        """Write the study config to a YAML file."""
        self.study_yaml_path.parent.mkdir(parents=True, exist_ok=True)
        with self.study_yaml_path.open("w") as file:
            yaml.dump(self.to_dict(), file, sort_keys=False)


    @classmethod
    def from_problems(  # noqa: C901, PLR0912, PLR0915
        cls,
        problems: Problem | list[Problem],
        seeds: Iterable[int] | None = None,
        num_seeds: int = 1,
        name: str | None = None,
        output_dir: Path | str | None = DEFAULT_STUDY_DIR,
        group_by: Literal["opt", "bench", "opt_bench", "seed", "mem"] | None = None,
        on_error: Literal["warn", "raise", "ignore"] = "warn"
    ) -> Study:
        """Create a Study object from a list of Problems.

        Args:
            name: The name of the study.

            output_dir: The main output directory where the hposuite studies are saved.

            problems: The list of problems to generate runs for.

            seeds: The seed or seeds to use for the experiment.

            num_seeds: The number of seeds to generate.

            group_by: The grouping to use for the runs dump.

            on_error: The method to handle errors while generating runs for the study.

        Returns:
            A Study object.
        """
        if not isinstance(problems, list):
            problems = [problems]


        match output_dir:
            case None:
                output_dir = DEFAULT_STUDY_DIR
            case str():
                output_dir = Path(output_dir)
            case Path():
                pass
            case _:
                raise TypeError(f"Invalid type for output_dir: {type(output_dir)}")


        if not seeds and not (num_seeds and num_seeds > 0):
            match on_error:
                case "raise":
                    raise ValueError("At least one seed or num_seeds must be provided")
                case "warn" | "ignore":
                    warnings.warn(
                        "At least one seed or num_seeds must be provided"
                        "Continuing with num_seeds=1",
                        stacklevel=2
                    )
                    num_seeds = 1
                case _:
                    raise TypeError(
                        f"Invalid value for on_error: {on_error}"
                    )

        # Generate seeds
        match seeds:
            case None:
                seeds = cls.generate_seeds(num_seeds)
            case Iterable():
                seeds = list(set(seeds))
            case int():
                seeds = [seeds]


        _budget = problems[0].budget.total
        _problems: list[Problem] = []
        for _problem in problems:
            try:
                if not isinstance(_problem, Problem):
                    raise TypeError(
                        f"Expected Problem or list[Problem], got {type(_problem)}"
                    )
                if _problem.budget.total != _budget:
                    raise ValueError("All problems must have the same budget")
                _problems.append(_problem)
            except Exception as e:
                match on_error:
                    case "raise":
                        raise e
                    case "ignore":
                        continue
                    case "warn":
                        warnings.warn(f"{e}\nTo ignore this, set `on_error='ignore'`", stacklevel=2)
                        continue


        #Generate runs from problems
        _runs_per_problem: Mapping[str, Run] = {}
        for _problem, _seed in product(_problems, seeds):
            try:
                _run = Run(
                        problem=_problem,
                        seed=_seed,
                    )
                if _run.name not in _runs_per_problem:
                    _runs_per_problem[_run.name] = _run
            except ValueError as e:
                match on_error:
                    case "raise":
                        raise e
                    case "ignore":
                        continue
                    case "warn":
                        warnings.warn(f"{e}\nTo ignore this, set `on_error='ignore'`", stacklevel=2)
                        continue
        _runs_per_problem = list(_runs_per_problem.values())

        logger.info(f"Generated {len(_runs_per_problem)} runs")

        return cls(
            name=name,
            output_dir=output_dir,
            experiments=_runs_per_problem,
            seeds=seeds,
            num_seeds=num_seeds,
            budget=_budget,
            group_by=group_by
        )


    @classmethod
    def generate_seeds(
        cls,
        num_seeds: int,
        offset: int = 0, # To offset number of seeds
    ) -> list[int]:
        """Generate a set of seeds using a Global Seed."""
        cls._rng = np.random.default_rng(GLOBAL_SEED)
        _num_seeds = num_seeds + offset
        _seeds = cls._rng.integers(0, 2 ** 32, size=_num_seeds)
        return _seeds[offset:].tolist()


    @classmethod
    def generate(  # noqa: C901, PLR0912, PLR0915
        cls,
        optimizers: (
            type[Optimizer]
            | OptWithHps
            | list[type[Optimizer]]
            | list[OptWithHps | type[Optimizer]]
        ),
        benchmarks: (
            BenchmarkDescription
            | BenchWith_Objs_Fids
            | list[BenchmarkDescription]
            | list[BenchWith_Objs_Fids | BenchmarkDescription]
        ),
        *,
        budget: BudgetType | int,
        seeds: Iterable[int] | None = None,
        num_seeds: int = 1,
        multi_objective_generation: Literal["mix_metric_cost", "metric_only"] = "mix_metric_cost",
        on_error: Literal["warn", "raise", "ignore"] = "warn",
        continuations: bool = True,
    ) -> list[Run]:
        """Generate a set of problems for the given optimizer and benchmark.

        If there is some incompatibility between the optimizer, the benchmark and the requested
        amount of objectives, fidelities or costs, a ValueError will be raised.

        Args:
            optimizers: The optimizer class to generate problems for.
                Can provide a single optimizer or a list of optimizers.
                If you wish to provide hyperparameters for the optimizer, provide a tuple with the
                optimizer.

            benchmarks: The benchmark to generate problems for.
                Can provide a single benchmark or a list of benchmarks.

            expdir: Which directory to store experiment results into.

            budget: The budget to use for the problems. Budget defaults to a n_trials budget
                where when multifidelty is enabled, fractional budget can be used and 1 is
                equivalent a full fidelity trial.

            seeds: The seed or seeds to use for the problems.

            num_seeds: The number of seeds to generate. Only used if seeds is None.

            multi_objective_generation: The method to generate multiple objectives.

            on_error: The method to handle errors.
                * "warn": Log a warning and continue.
                * "raise": Raise an error.
                * "ignore": Ignore the error and continue.

            continuations: Whether to use continuations for the run.

        Returns:
            A list of Run objects.

        Note:
            `Study.generate()` only generates the Run objects.
            It does not set absolute paths or write yaml files.
        """
        # Generate seeds
        match seeds:
            case None:
                seeds = cls.generate_seeds(num_seeds)
            case Iterable():
                pass
            case int():
                seeds = [seeds]

        _optimizers: list[OptWithHps]
        match optimizers:
            case tuple():
                _opt, hps = optimizers
                _optimizers = [(_opt, hps)]
            case list():
                _optimizers = [o if isinstance(o, tuple) else (o, {}) for o in optimizers]
            case type():
                _optimizers = [(optimizers, {})]
            case _:
                raise TypeError(
                    "Expected Optimizer or list[Optimizer] or tuple[Optimizer, dict] or "
                    f"list[tuple[Optimizer, dict]], got {type(optimizers)}"
                )

        _benchmarks: list[BenchWith_Objs_Fids]
        match benchmarks:
            case tuple():
                _bench, objsfids = benchmarks
                _benchmarks = [(_bench, objsfids)]
            case list():
                _benchmarks = [b if isinstance(b, tuple) else (b, {}) for b in benchmarks]
            case BenchmarkDescription():
                _benchmarks = [(benchmarks, {})]
            case _:
                raise TypeError(
                    "Expected BenchmarkDescription or list[BenchmarkDescription] or "
                    "tuple[BenchmarkDescription, dict] or list[tuple[BenchmarkDescription, dict]],"
                    f" got {type(benchmarks)}"
                )

        _problems: list[Problem] = []
        for (opt, hps), (bench, objs_fids) in product(_optimizers, _benchmarks):
            try:

                objectives: int | str | list[str]
                fidelities: int | str | list[str] | None
                costs: int | str | list[str] | None

                objectives = objs_fids.get("objectives", 1)
                if isinstance(objectives, list) and len(objectives) == 1:
                    objectives = objectives[0]

                fidelities = objs_fids.get("fidelities", None)
                if isinstance(fidelities, list) and len(fidelities) == 1:
                    fidelities = fidelities[0]

                costs = objs_fids.get("costs", None)
                if isinstance(costs, list) and len(costs) == 1:
                    costs = costs[0]

                priors = objs_fids.get("priors")

                match fidelities, bench.fidelities:
                    case None, None:
                        fidelities = None
                    case None, _:
                        match opt.support.fidelities[0]:
                            case "single":
                                fidelities = 1
                            case "many":
                                fidelities = len(bench.fidelities)
                            case None:
                                fidelities = None
                            case _:
                                raise ValueError("Invalid fidelity support")
                    case str() | int() | list(), _:
                        match opt.support.fidelities[0]:
                            case str():
                                pass
                            case None:
                                fidelities = None
                            case _:
                                raise ValueError("Invalid fidelity support")
                    case _:
                        raise ValueError(
                            f"Invalid fidelity type: {type(fidelities)}. "
                            f"Expected None, str, int or list"
                        )

                if priors:
                    priors = dict_to_configpriors(priors)


                _problem = Problem.problem(
                    optimizer=opt,
                    optimizer_hyperparameters=hps,
                    benchmark=bench,
                    objectives=objectives,
                    budget=budget,
                    fidelities=fidelities,
                    costs=costs,
                    multi_objective_generation=multi_objective_generation,
                    continuations=continuations,
                    priors=priors,
                )
                _problems.append(_problem)
            except ValueError as e:
                match on_error:
                    case "raise":
                        raise e
                    case "ignore":
                        continue
                    case "warn":
                        warnings.warn(f"{e}\nTo ignore this, set `on_error='ignore'`", stacklevel=2)
                        continue


        # NOTE: REDUNDANCY CHECK: _runs_per_problem is a dict now to avoid duplicate runs esp.
        # for eg. in case a BO Opt being used in combination with a
        # MF Opt on a benchmark with multiple fidelities -> explicitly adding different fidelities
        # in multiple benchmark instances would create redundant problems if BO Opts are present.
        _runs_per_problem: Mapping[str, Run] = {}
        for _problem, _seed in product(_problems, seeds):
            try:
                _run = Run(
                        problem=_problem,
                        seed=_seed,
                    )
                if _run.name not in _runs_per_problem:
                    _runs_per_problem[_run.name] = _run
            except ValueError as e:
                match on_error:
                    case "raise":
                        raise e
                    case "ignore":
                        continue
                    case "warn":
                        warnings.warn(f"{e}\nTo ignore this, set `on_error='ignore'`", stacklevel=2)
                        continue
        _runs_per_problem = list(_runs_per_problem.values())

        logger.info(f"Generated {len(_runs_per_problem)} runs")

        return _runs_per_problem


    def _group_by(  # noqa: C901
        self,
        group_by: Literal["opt", "bench", "opt_bench", "seed", "mem"] | None,
    ) -> Mapping[str, list[Run]]:
        """Group the runs by the specified group."""
        _grouped_runs = {}
        filtered_runs = {}

        # To avoid duplicate run dumps because we do not support having
        # objectives, fidelities and priors in the hpoglue CLI run command
        for run in self.experiments:
            _run_hash = ".".join(
                [
                    run.optimizer.name,
                    run.benchmark.name,
                    str(run.seed),
                    f"{run.mem_req_mb}mb",
                ]
            )
            if _run_hash not in filtered_runs:
                filtered_runs[_run_hash] = run

        if group_by is None:
            return {"all": list(filtered_runs.values())}

        for run in filtered_runs.values():
            key = ""
            match group_by:
                case "opt":
                    key = run.optimizer.name
                case "bench":
                    key = run.benchmark.name
                case "opt_bench":
                    key = f"{run.optimizer.name}_{run.benchmark.name}"
                case "seed":
                    key = str(run.seed)
                case "mem":
                    key = f"{run.mem_req_mb}mb"
                case _:
                    raise ValueError(f"Invalid group_by: {group_by}")

            if key not in _grouped_runs:
                _grouped_runs[key] = []
            _grouped_runs[key].append(run)

        return _grouped_runs

    def _dump_runs(
        self,
        group_by: Literal["opt", "bench", "opt_bench", "seed", "mem"] | None,
        exp_dir: Path,
    ) -> None:
        """Dump the grouped runs into separate files."""
        grouped_runs = self._group_by(group_by)
        for key, runs in grouped_runs.items():
            with (exp_dir / f"dump_{key}.txt").open("w") as f:
                for run in runs:
                    f.write(
                        f"python -m hpoglue"
                        f" --optimizers {run.optimizer.name}"
                        f" --benchmarks {run.benchmark.name}"
                        f" --seeds {run.seed}"
                        f" --budget {run.problem.budget.total}"
                    )
                    f.write("\n")
            logger.info(f"Dumped experiments to {(exp_dir / f'dump_{key}.txt').absolute()}")


    def optimize(  # noqa: C901, PLR0912
        self,
        *,
        exec_type: Literal["sequential", "parallel"] = "sequential",
        add_seeds: Iterable[int] | int | None = None,
        add_num_seeds: int | None = None,
        overwrite: bool = False,
        continuations: bool = True,
        auto_env_handling: bool = False,
    ) -> None:
        """Execute multiple atomic runs using a list of Optimizers and a list of Benchmarks.

        Args:
            exec_type: The type of execution to use.
            Supported types are "sequential", "parallel" and "dump".

            add_seeds: The seed or seeds to add to the study.

            add_num_seeds: The number of seeds to generate and add to the study.

            NOTE: Only one of `add_seeds` and `add_num_seeds` can be provided.

            group_by: The grouping to use for the runs dump.
            Supported types are "opt", "bench", "opt_bench", "seed", and "mem"
            Only used when `exec_type` is "dump" for multiple runs.

            overwrite: Whether to overwrite existing results.

            continuations: Whether to calculate continuations cost.
                Note: Only works for Multi-fidelity Optimizers.

            on_error: The method to handle errors.
                    * "warn": Log a warning and continue.
                    * "raise": Raise an error.
                    * "ignore": Ignore the error and continue.

            auto_env_handling: Whether to automatically create and use isolated run environments.

        """
        if add_seeds is not None and add_num_seeds is not None:
            logger.warning(
                "Cannot provide both `add_seeds` and `add_num_seeds`!"
                "Using only `add_seeds` and ignoring `add_num_seeds`"
            )
            add_num_seeds = None
        if isinstance(add_seeds, int):
            add_seeds = [add_seeds]

        _seeds: list[int] = []

        match add_seeds, add_num_seeds:
            case None, None:
                pass
            case Iterable(), None:
                add_seeds = list(set(add_seeds))
                for seed in add_seeds:
                    if seed not in self.seeds:
                        _seeds.append(seed)
            case None, int():
                _num_seeds = add_num_seeds
                offset = 0
                while _num_seeds > 0:
                    new_seeds = [
                        s for s in Study.generate_seeds(_num_seeds, offset=offset)
                        if s not in self.seeds
                    ]
                    _seeds.extend(new_seeds)
                    _num_seeds -= len(_seeds)
                    offset += _num_seeds + len(new_seeds)
            case _:
                raise ValueError(
                    "Invalid combination of types for `add_seeds` and `add_num_seeds`"
                    "Expected (Iterable[int] | int | None, int | None),"
                    f"got ({type(add_seeds)}, {type(add_num_seeds)})"
                )


        if _seeds:
            self._update_study(new_seeds=_seeds)

        if overwrite:
            logger.info("Overwrite flag is set to True. Existing results will be overwritten!")

        if (len(self.experiments) > 1):
            match exec_type:
                case "sequential":
                    logger.info(f"Running {len(self.experiments)} experiments sequentially")
                    for i, run in enumerate(self.experiments, start=1):
                        if auto_env_handling:
                            run.create_env(hposuite=f"-e {HPOSUITE_EDITABLE}")
                        logger.info(f"Running experiment {i}/{len(self.experiments)}")
                        run.run(
                            continuations=continuations,
                            overwrite=overwrite,
                            progress_bar=False
                        )
                case "parallel":
                    raise NotImplementedError("Parallel execution not implemented yet!")
                case _:
                    raise ValueError(f"Invalid exceution type: {exec_type}")
        else:
            run = self.experiments[0]
            if auto_env_handling:
                run.create_env(hposuite=f"-e {Path.cwd()}")
            logger.info("Running single experiment")
            run.run(
                continuations=continuations,
                overwrite=overwrite,
                progress_bar=False,
            )
        logger.info(f"Completed study with {len(self.experiments)} runs")


def create_study(  # noqa: C901, PLR0912, PLR0915
    *,
    name: str | None = None,
    output_dir: str| Path | None = DEFAULT_STUDY_DIR,
    optimizers: (
        str
        | tuple[str, Mapping[str, Any]]
        | type[Optimizer]
        | OptWithHps
        | list[tuple[str, Mapping[str, Any]]]
        | list[str]
        | list[OptWithHps]
        | list[type[Optimizer]]
    ),
    benchmarks: (
        str
        | BenchmarkDescription
        | FunctionalBenchmark
        | tuple[str, Mapping[str, Any]]
        | BenchWith_Objs_Fids # tuple[BenchmarkDescription, Mapping[str, Any]]
        | tuple[FunctionalBenchmark, Mapping[str, Any]]
        | list[str]
        | list[BenchmarkDescription]
        | list[FunctionalBenchmark]
        | list[tuple[str, Mapping[str, Any]]]
        | list[BenchWith_Objs_Fids]   # list[tuple[BenchmarkDescription, Mapping[str, Any]]]
        | list[tuple[FunctionalBenchmark, Mapping[str, Any]]]
    ),
    seeds: Iterable[int] | int | None = None,
    num_seeds: int = 1,
    budget: int = 50,
    group_by: Literal["opt", "bench", "opt_bench", "seed", "mem"] | None = None,
    on_error: Literal["warn", "raise", "ignore"] = "warn",
) -> Study:
    """Create a Study object.

    Args:
        name: The name of the study.

        output_dir: The main output directory where the hposuite studies are saved.

        optimizers: The list of optimizers to use.
                    Usage: [
                        (
                            optimizer: str | type[Optimizer],
                            {
                                "hp_name": hp_value
                            }
                        )
                    ]

        benchmarks: The list of benchmarks to use.
                    Usage: [
                        (
                            benchmark: str | BenchmarkDescription | FunctionalBenchmark,
                            {
                                "objectives": [list of objectives] | number of objectives,
                                "fidelities": [list of fidelities] | number of fidelities | None,
                                "costs": [list of costs] | number of costs | None,
                                "priors": {
                                    "objective": Prior
                                }
                            }
                        )
                    ]

        seeds: The seed or seeds to use for the experiment.

        num_seeds: The number of seeds to generate.

        budget: The budget for the experiment.

        group_by: The grouping to use for the runs dump.

        on_error: The method to handle errors while generating runs for the study.

    Returns:
        A Study object.
    """
    # TODO: Add redundancy checks for repeated optimizers and benchmarks with same configs
    match output_dir:
        case None:
            output_dir = DEFAULT_STUDY_DIR
        case str():
            output_dir = Path(output_dir)
        case Path():
            pass
        case _:
            raise TypeError(f"Invalid type for output_dir: {type(output_dir)}")

    assert optimizers, "At least one optimizer must be provided!"
    if not isinstance(optimizers, list):
        optimizers = [optimizers]

    assert benchmarks, "At least one benchmark must be provided!"
    if not isinstance(benchmarks, list):
        benchmarks = [benchmarks]


    if not seeds and not (num_seeds and num_seeds > 0):
        match on_error:
            case "raise":
                raise ValueError("At least one seed or num_seeds must be provided")
            case "warn" | "ignore":
                warnings.warn(
                    "At least one seed or num_seeds must be provided"
                    "Continuing with num_seeds=1",
                    stacklevel=2
                )
                num_seeds = 1
            case _:
                raise TypeError(
                    f"Invalid value for on_error: {on_error}"
                )


    _optimizers: list[OptWithHps] = []
    for optimizer in optimizers:
        match optimizer:
            case str():
                assert optimizer in OPTIMIZERS, (
                    f"Optimizer must be one of {OPTIMIZERS.keys()}\n"
                    f"Found {optimizer}"
                )
                _optimizers.append(OPTIMIZERS[optimizer])
            case tuple():
                opt, hps = optimizer
                match opt:
                    case str():
                        assert opt in OPTIMIZERS, (
                            f"Optimizer must be one of {OPTIMIZERS.keys()}\n"
                            f"Found {opt}"
                        )
                        _optimizers.append((OPTIMIZERS[opt], hps))
                    case type():
                        _optimizers.append((opt, hps))
                    case _:
                        raise TypeError(f"Unknown Optimizer type {type(opt)}")
            case type():
                _optimizers.append(optimizer)
            case _:
                raise TypeError(f"Unknown Optimizer type {type(optimizers)}")


    _benchmarks: list[BenchWith_Objs_Fids] = []
    for benchmark in benchmarks:
        match benchmark:
            case str():
                assert benchmark in BENCHMARKS, (
                    f"Benchmark must be one of {BENCHMARKS.keys()}\n"
                    f"Found {benchmark}"
                )
                if not isinstance(BENCHMARKS[benchmark], FunctionalBenchmark):
                    _benchmarks.append(BENCHMARKS[benchmark])
                else:
                    _benchmarks.append(BENCHMARKS[benchmark].desc)
            case BenchmarkDescription():
                _benchmarks.append(benchmark)
            case FunctionalBenchmark():
                _benchmarks.append(benchmark.desc)
            case tuple():
                bench, bench_hps = benchmark
                match bench:
                    case str():
                        assert bench in BENCHMARKS, (
                            f"Benchmark must be one of {BENCHMARKS.keys()}\n"
                            f"Found {bench}"
                        )
                        if not isinstance(BENCHMARKS[bench], FunctionalBenchmark):
                            _benchmarks.append((BENCHMARKS[bench], bench_hps))
                        else:
                            _benchmarks.append((BENCHMARKS[bench].desc, bench_hps))
                    case BenchmarkDescription():
                        _benchmarks.append((bench, bench_hps))
                    case FunctionalBenchmark():
                        _benchmarks.append((bench.desc, bench_hps))
                    case _:
                        raise TypeError(f"Unknown Benchmark type {type(benchmark)}")
            case _:
                raise TypeError(f"Unknown Benchmark type {type(benchmarks)}")

    if isinstance(seeds, Iterable):
        seeds = list(set(seeds))

    experiments = Study.generate(
        optimizers=_optimizers,
        benchmarks=_benchmarks,
        budget=budget,
        seeds=seeds,
        num_seeds=num_seeds,
        on_error=on_error,
    )

    return Study(
        name=name,
        output_dir=output_dir,
        experiments=experiments,
        seeds=seeds,
        num_seeds=num_seeds,
        budget=budget,
        group_by=group_by,
    )


from __future__ import annotations

import logging
from collections.abc import Mapping
from pathlib import Path
from typing import TYPE_CHECKING
from typing_extensions import Any

import ConfigSpace as CS  # noqa: N817
import nevergrad as ng
import numpy as np
from hpoglue import Config, Optimizer, Problem, Query
from hpoglue.env import Env
from nevergrad.parametrization import parameter

if TYPE_CHECKING:
    from hpoglue import Result
    from nevergrad.optimization.base import (
        ConfiguredOptimizer as ConfNGOptimizer,
        Optimizer as NGOptimizer,
    )

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

opt_list = sorted(ng.optimizers.registry.keys())
ext_opts = {
    "Hyperopt": ng.optimization.optimizerlib.NGOpt13,
    "CMA-ES": ng.optimization.optimizerlib.ParametrizedCMA,
    "bayes_opt": ng.optimization.optimizerlib.ParametrizedBO,
    "DE": ng.families.DifferentialEvolution,
    "EvolutionStrategy": ng.families.EvolutionStrategy,
}


class NevergradOptimizer(Optimizer):
    """The Nevergrad Optimizer."""

    name = "Nevergrad"

    env = Env(
        name="nevergrad-1.0.5",
        python_version="3.10",
        requirements=("nevergrad==1.0.5",),
    )

    support = Problem.Support(
        fidelities=(None,),
        objectives=("single", "many"),
        cost_awareness=(None,),
        tabular=False,
    )

    mem_req_mb = 1024

    def __init__(
        self,
        *,
        problem: Problem,
        seed: int,
        working_directory: Path,
        optimizer_name: str = "NGOpt",
    ) -> None:
        """Create a Nevergrad Optimizer instance for a given Problem."""
        self._parametrization: dict[str, parameter.Parameter]
        config_space = problem.config_space
        match config_space:
            case CS.ConfigurationSpace():
                self._parametrization = _configspace_to_nevergrad_space(config_space)
            case list():
                raise NotImplementedError("# TODO: Tabular not yet implemented for Nevergrad!")
            case _:
                raise TypeError("Config space must be a list or a ConfigurationSpace!")

        self.seed = seed
        self.problem = problem
        self.working_directory = working_directory

        if optimizer_name not in list(ext_opts.keys()) + opt_list:
            raise ValueError(f"Unknown optimizer: {optimizer_name}!")

        self.optimizer: NGOptimizer | ConfNGOptimizer
        match problem.objectives:
            case tuple() | Mapping():
                if optimizer_name in ext_opts:
                    if optimizer_name == "Hyperopt":
                        ng_opt = ext_opts["Hyperopt"]
                    else:
                        ng_opt = ext_opts[optimizer_name]()
                    self.optimizer = ng_opt(
                        parametrization=self._parametrization,
                        budget=self.problem.budget.total, # TODO: Check this
                        num_workers=1,
                    )
                else:
                    self.optimizer = ng.optimizers.registry[optimizer_name](
                        parametrization=self._parametrization,
                        budget=self.problem.budget.total, # TODO: Check this
                        num_workers=1,
                    )
                self.optimizer.parametrization.random_state = np.random.RandomState(seed)
            case _:
                raise ValueError("Objective must be a string or a list of strings!")

        logger.info(f"Initialized Nevergrad Optimizer with {self.optimizer.name}!")

        self.history: dict[str, tuple[ng.p.Parameter, float | list[float] | None]] = {}
        self.counter = 0


    def ask(self) -> Query:
        match self.problem.fidelities:
            case None:
                config: parameter.Parameter = self.optimizer.ask()
                name = f"{self.counter}_{config.value}_{self.seed}"
                self.history[name] = (config, None)
                self.counter += 1
                return Query(
                    config=Config(
                        config_id=name,
                        values=config.value),
                        # allow_inactive_with_values=True,
                    fidelity=None,
                    optimizer_info=config,
                )
            case tuple():
                # TODO(eddiebergman): Not sure if just using
                # trial.number is enough in MF setting
                raise NotImplementedError("# TODO: Fidelity-aware not available in Nevergrad!")
            case Mapping():
                raise NotImplementedError("# TODO: Fidelity-aware not available in Nevergad!")
            case _:
                raise TypeError("Fidelity must be None or a tuple!")


    def tell(self, result: Result) -> None:
        match self.problem.objectives:
            case (name, _):
                _values = result.values[name]
            case Mapping():
                _values = [result.values[key] for key in self.problem.objectives]
            case _:
                raise TypeError("Objective must be a string or a list of strings!")

        match self.problem.costs:
            case None:
                pass
            case tuple():
                raise NotImplementedError("# TODO: Cost-aware not yet implemented for Nevergrad!")
            case Mapping():
                raise NotImplementedError("# TODO: Cost-aware not yet implemented for Nevergrad!")
            case _:
                raise TypeError("Cost must be None or a mapping!")

        assert isinstance(result.query.optimizer_info, parameter.Parameter)
        self.optimizer.tell(
            result.query.optimizer_info,
            _values,
        )


def _configspace_to_nevergrad_space(  # noqa: C901, PLR0912
    config_space: CS.ConfigurationSpace,
) -> dict[str, ng.p.Instrumentation]:

    if len(config_space.conditions) > 0:
        raise NotImplementedError("Conditions are not yet supported!")

    if len(config_space.forbidden_clauses) > 0:
        raise NotImplementedError("Forbiddens are not yet supported!")

    ng_space = ng.p.Dict()
    for hp in list(config_space.values()):
        match hp:
            case CS.UniformIntegerHyperparameter():
                if hp.log:
                    ng_space[hp.name] = ng.p.Log(lower=hp.lower, upper=hp.upper).set_integer_casting()
                else:
                    ng_space[hp.name] = ng.p.Scalar(lower=hp.lower, upper=hp.upper).set_integer_casting()
            case CS.UniformFloatHyperparameter():
                if hp.log:
                    ng_space[hp.name] = ng.p.Log(lower=hp.lower, upper=hp.upper)
                else:
                    ng_space[hp.name] = ng.p.Scalar(lower=hp.lower, upper=hp.upper)
            case CS.CategoricalHyperparameter():
                if hp.weights is not None:
                    raise NotImplementedError("Weights on categoricals are not yet supported!")
                ng_space[hp.name] = ng.p.Choice(hp.choices)
            case CS.Constant():
                ng_space[hp.name] = ng.p.Choice([hp.value])
            case CS.OrdinalHyperparameter():
                ng_space[hp.name] = ng.p.TransitionChoice(hp.sequence)
            case _:
                raise ValueError("Unrecognized type of hyperparameter in ConfigSpace!")

    return ng_space



from __future__ import annotations

import logging
import shutil
import subprocess
import traceback
from collections.abc import Hashable, Iterable, Mapping, Sequence
from dataclasses import dataclass, field
from enum import Enum
from pathlib import Path
from typing import Any, Literal, TypeAlias, TypeVar

import numpy as np
import pandas as pd
import yaml
from hpoglue import BenchmarkDescription, Config, Optimizer, Problem, Query, Result
from hpoglue.budget import CostBudget, TrialBudget
from hpoglue.dataframe_utils import reduce_dtypes
from hpoglue.env import (
    Env,
    Venv,
)

from hposuite.utils import HPOSUITE_PYPI, get_current_installed_hposuite_version

OptWithHps: TypeAlias = tuple[type[Optimizer], Mapping[str, Any]]

T = TypeVar("T", bound=Hashable)


logger = logging.getLogger(__name__)

GLOBAL_SEED = 42


def _try_delete_if_exists(path: Path) -> None:
    if path.exists():
        try:
            if path.is_dir():
                shutil.rmtree(path)
            else:
                path.unlink()

        except Exception as e:
            logger.exception(e)
            logger.error(f"Error deleting {path}: {e}")


@dataclass
class Run:
    """A run of a benchmark."""

    problem: Problem
    """The problem that was run."""

    seed: int
    """The seed used for the run."""

    optimizer: type[Optimizer] = field(init=False)
    """The optimizer to use for the Problem"""

    expdir: Path = field(init=False)
    """Default main directory to use for the run's Study."""

    optimizer_hyperparameters: Mapping[str, Any] = field(default_factory=dict, init=False)
    """The hyperparameters to use for the optimizer"""

    benchmark: BenchmarkDescription = field(init=False)
    """The benchmark that the Problem was run on."""

    env: Env = field(init=False)
    """The environment to setup the optimizer in for `isolated` mode."""

    working_dir: Path = field(init=False)
    """The working directory for the run."""

    complete_flag: Path = field(init=False)
    """The flag to indicate the run is complete."""

    error_file: Path = field(init=False)
    """The file to store the error traceback if the run crashed."""

    running_flag: Path = field(init=False)
    """The flag to indicate the run is currently running."""

    queue_flag: Path = field(init=False)
    """The flag to indicate the run is queued."""

    df_path: Path = field(init=False)
    """The path to the dataframe for the run."""

    venv_requirements_file: Path = field(init=False)
    """The path to the requirements file for the run."""

    post_install_steps: Path = field(init=False)
    """The path to the post install steps file for the run."""

    run_yaml_path: Path = field(init=False)
    """The path to the run yaml file for the run."""

    env_description_file: Path = field(init=False)
    """The path to the environment description file for the run."""

    env_path: Path = field(init=False)
    """The path to the environment for the run."""

    mem_req_mb: int = field(init=False)
    """The memory requirement for the run in mb.
    Calculated as the sum of the memory requirements of the optimizer and the benchmark.
    """

    def __post_init__(self) -> None:
        self.benchmark = self.problem.benchmark
        self.optimizer = self.problem.optimizer
        self.optimizer_hyperparameters = self.problem.optimizer_hyperparameters
        self.mem_req_mb = self.problem.mem_req_mb

        name_parts: list[str] = [
            self.problem.name,
            f"seed={self.seed}",
        ]
        self.name = ".".join(name_parts)

        match self.benchmark.env, self.optimizer.env:
            case (None, None):
                self.env = Env.empty()
            case (None, Env()):
                self.env = self.optimizer.env
            case (Env(), None):
                self.env = self.benchmark.env
            case (Env(), Env()):
                self.env = Env.merge(self.benchmark.env, self.optimizer.env)
            case _:
                raise ValueError("Invalid combination of benchmark and optimizer environments")


    def _set_paths(self, expdir: Path | str) -> None:
        if isinstance(expdir, str):
            expdir = Path(expdir)
        self.expdir = expdir
        self.working_dir = self.expdir.absolute().resolve() / self.name
        self.complete_flag = self.working_dir / "complete.flag"
        self.error_file = self.working_dir / "error.txt"
        self.running_flag = self.working_dir / "running.flag"
        self.df_path = self.working_dir / f"{self.name}.parquet"
        self.venv_requirements_file = self.working_dir / "venv_requirements.txt"
        self.queue_flag = self.working_dir / "queue.flag"
        self.requirements_ran_with_file = self.working_dir / "requirements_ran_with.txt"
        self.env_description_file = self.working_dir / "env.yaml"
        self.post_install_steps = self.working_dir / "venv_post_install.sh"
        self.run_yaml_path = self.working_dir / "run_config.yaml"
        self.env_path = self.expdir / "envs" / self.env.identifier
        self.env_error_file = self.env_path / "error.txt"
        self.env_complete_file = self.env_path / "complete.txt"


    @property
    def venv(self) -> Venv:
        """Creates and returns an hpoglue Venv object using the specified environment path."""
        return Venv(self.env_path)

    @property
    def conda(self) -> Venv:
        """Creates and returns an hpoglue Conda object using the specified environment path."""
        raise NotImplementedError("Conda not implemented yet.")

    def run(  # noqa: C901, PLR0912
        self,
        *,
        continuations: bool = True,
        on_error: Literal["raise", "continue"] = "raise",
        overwrite: Run.State | str | Sequence[Run.State | str] | bool = False,
        progress_bar: bool = False,
        auto_env_handling: bool = False,
    ) -> Report:
        """Run the Run.

        Args:
            on_error: How to handle errors. In any case, the error will be written
                into the [`working_dir`][hpoglue.run.Run.working_dir]
                of the problem.

                * If "raise", raise an error.
                * If "continue", log the error and continue.

            overwrite: What to overwrite.

                * If a single value, overwrites problem in that state,
                * If a list of states, overwrites any problem in one of those
                 states.
                * If `True`, overwrite problems in all states.
                * If `False`, don't overwrite any problems.

            progress_bar: Whether to show a progress bar.

            continuations: Whether to use continuations for the run.

            auto_env_handling: Whether to automatically use the created environment for this run.
        """
        from hpoglue._run import _run

        if on_error not in ("raise", "continue"):
            raise ValueError(f"Invalid value for `on_error`: {on_error}")

        overwrites = Run.State.collect(overwrite)

        state = self.state()
        if state in overwrites:
            logger.info(f"Overwriting {self.name} in `{state=}` at {self.working_dir}.")
            self.set_state(Run.State.PENDING)

        if self.df_path.exists():
            logger.info(f"Loading results for {self.name} from {self.working_dir}")
            return Run.Report.from_df(
                df=pd.read_parquet(self.df_path),
                run=self,
            )

        """ TODO
        if self.working_dir.exists():
            raise RuntimeError(
                "The optimizer ran before but no dataframe of results was found at "
                f"{self.df_path}. Set `overwrite=[{state}]` to rerun problems in this state"
            )
        """
        self.set_state(Run.State.PENDING)
        _hist: list[Result] = []
        try:
            match continuations, self.problem.continuations:
                case True, True:
                    pass
                case False, False:
                    pass
                case True, False:
                    if "single" not in self.optimizer.support.fidelities:
                        logger.warning(
                            f"Continuations are not supported for the optimizer: {self.optimizer}."
                            " Ignoring continuations=True."
                        )
                    else:
                        self.problem.continuations = True
                case False, True:
                    self.problem.continuations = False
                case _:
                    raise RuntimeError("continuations expects a bool value!")

            self.set_state(self.State.RUNNING, auto_env_handling=auto_env_handling)
            _hist = _run(
                problem=self.problem,
                seed=self.seed,
                run_name=self.name,
                on_error="raise",
                progress_bar=progress_bar,
            )
        except Exception as e:
            self.set_state(Run.State.CRASHED, err_tb=(e, traceback.format_exc()))
            logger.exception(e)
            logger.error(f"Error in Run {self.name}: {e}")
            match on_error:
                case "raise":
                    raise e
                case "continue":
                    raise NotImplementedError("Continue not yet implemented!") from e
                case _:
                    raise RuntimeError(f"Invalid value for `on_error`: {on_error}") from e
        logger.info(f"Results dumped at {self.df_path.absolute()}")
        return self.post_process(history=_hist)


    def post_process(
        self,
        *,
        history: list[Result],
    ) -> Report:
        """Post process the run.

        Args:
            history: The history of the run.

        Returns:
            The Report of the run.
        """
        report = Run.Report(
            run=self,
            results=history,
        )
        self.set_state(Run.State.COMPLETE, df=report.df())
        return report


    def create_env(  # noqa: C901, PLR0912, PLR0915
        self,
        *,
        how: Literal["venv", "conda"] = "venv",
        hposuite: Literal["current_version"] | str,
    ) -> None:
        """Set up the isolation for the experiment."""
        if hposuite == "current_version":
            raise NotImplementedError("Not implemented yet.")


        match hposuite:
            case "current_version":
                _version = get_current_installed_hposuite_version()
                req = f"{HPOSUITE_PYPI}=={_version}"
            case str():
                req = hposuite
            case _:
                raise ValueError(f"Invalid value for `hposuite`: {hposuite}")

        requirements = [req, *self.env.requirements]

        logger.info(f"Installing deps: {self.env.identifier}")
        self.working_dir.mkdir(parents=True, exist_ok=True)
        with self.venv_requirements_file.open("w") as f:
            f.write("\n".join(requirements))


        pi_flag = False

        if self.env_complete_file.exists():
            logger.info(f"Environment already installed: {self.env.identifier}")
            if self.env.post_install:
                with self.env_complete_file.open("r") as f:
                    if "post_install" in f.read():
                        logger.info("Post install already ran.")
                        return
            else:
                return

        elif self.env_error_file.exists() or self.env_path.exists():
            logger.info(f"Cleaning up incomplete environment: {self.env.identifier}")
            shutil.rmtree(self.env_path)

        self.env_path.parent.mkdir(parents=True, exist_ok=True)

        env_dict = self.env.to_dict()
        env_dict.update({"env_path": str(self.env_path), "hposuite_source": req})

        logger.info(f"Installing env: {self.env.identifier}")
        match how:
            case "venv":
                try:
                    logger.info(f"Creating environment {self.env.identifier} at {self.env_path}")
                    self.venv.create(
                        path=self.env_path,
                        python_version=self.env.python_version,
                        requirements_file=self.venv_requirements_file,
                        exists_ok=False,
                    )
                    try:
                        if self.env_error_file.exists() and self.env_error_file.is_file():
                            self.env_error_file.unlink()
                    except Exception as e:
                        logger.error(f"Error removing error file: {e}")
                        raise e
                    self.env_complete_file.touch()
                except Exception as e:
                    with self.env_error_file.open("w") as f:
                        f.write(str(e))
                    raise e
                try:
                    if self.env.post_install and not pi_flag:
                        logger.info(f"Running post install for {self.env.identifier}")
                        with self.post_install_steps.open("w") as f:
                            f.write("\n".join(self.env.post_install))
                        self.venv.run(self.env.post_install)
                        with self.env_complete_file.open("w") as f:
                            f.write("post_install")
                except Exception as e:
                    logger.error(f"Error running post install steps: {e}")
                    raise e
            case "conda":
                raise NotImplementedError("Conda not implemented yet.")
            case _:
                raise ValueError(f"Invalid value for `how`: {how}")

    def to_dict(self) -> dict[str, Any]:
        """Convert the Run instance to a dictionary representation.

        Returns:
            A dictionary containing the object's data with keys:
                - "problem": The dictionary representation of the hpoglue Problem.
                - "seed": The seed value.
                - "expdir": The string representation of the study directory.
        """
        _problem = self.problem.to_dict()
        _problem["priors"] = list(_problem["priors"]) if _problem["priors"] else None
        return {
            "problem": _problem,
            "seed": self.seed,
            "expdir": str(self.expdir),
        }

    @classmethod
    def from_yaml(cls, path: Path) -> Run:
        """Create a Run instance from a YAML file."""
        if isinstance(path, str):
            path = Path(path)
        with path.open("r") as file:
            return Run.from_dict(yaml.safe_load(file))

    def write_yaml(self) -> None:
        """Writes the current object's data to a YAML file."""
        self.run_yaml_path.parent.mkdir(parents=True, exist_ok=True)
        with self.run_yaml_path.open("w") as file:
            yaml.dump(self.to_dict(), file, sort_keys=False)

    @classmethod
    def from_dict(cls, data) -> Run:
        """Create a Run instance from a dictionary.

        Args:
            data: A dictionary containing the data to create a Run instance.
                - "problem": An hpoglue Problem instance saved as a dictionary.
                - "seed": The seed value for the Run instance.

        Returns:
            A new instance of Run created from the provided dictionary.
        """
        from hposuite.utils import GlueWrapperFunctions
        run = Run(
            problem=GlueWrapperFunctions.problem_from_dict(data=data["problem"]),
            seed=data["seed"],
        )
        run._set_paths(expdir=Path(data["expdir"]))
        return run

    def state(self) -> Run.State:
        """Return the state of the run.

        Args:
            run: The run to get the state for.
        """
        if self.complete_flag.exists():
            return Run.State.COMPLETE

        if self.error_file.exists():
            return Run.State.CRASHED

        if self.running_flag.exists():
            return Run.State.RUNNING

        if self.queue_flag.exists():
            return Run.State.QUEUED

        return Run.State.PENDING

    def set_state(  # noqa: C901, PLR0912
        self,
        state: Run.State,
        *,
        df: pd.DataFrame | None = None,
        err_tb: tuple[Exception, str] | None = None,
        auto_env_handling: bool = False,
    ) -> None:
        """Set the run to a certain state.

        Args:
            state: The state to set the problem to.
            df: Optional dataframe to save if setting to [`Run.State.COMPLETE`].
            err_tb: Optional error traceback to save if setting to [`Run.State.CRASHED`].
            auto_env_handling: Whether to automatically create the environment for this run.
        """
        _flags = (self.complete_flag, self.error_file, self.running_flag, self.queue_flag)
        match state:
            case Run.State.PENDING:
                for _file in (*_flags, self.df_path, self.requirements_ran_with_file):
                    _try_delete_if_exists(_file)

                with self.run_yaml_path.open("w") as f:
                    yaml.dump(self.to_dict(), f, sort_keys=False)

            case Run.State.QUEUED:
                for _file in (*_flags, self.df_path, self.requirements_ran_with_file):
                    _try_delete_if_exists(_file)

                self.queue_flag.touch()

            case Run.State.RUNNING:
                for _file in (*_flags, self.df_path, self.requirements_ran_with_file):
                    _try_delete_if_exists(_file)

                self.working_dir.mkdir(parents=True, exist_ok=True)
                self.df_path.parent.mkdir(parents=True, exist_ok=True)

                if auto_env_handling:
                    lines = subprocess.run(  # noqa: S603
                        [self.venv.pip, "freeze"],
                        check=True,
                        capture_output=True,
                        text=True,
                    )
                    with self.requirements_ran_with_file.open("w") as f:
                        f.write(lines.stdout)

                self.running_flag.touch()

            case Run.State.CRASHED:
                for _file in (*_flags, self.df_path):
                    _try_delete_if_exists(_file)

                with self.error_file.open("w") as f:
                    if err_tb is None:
                        f.write("None")
                    else:
                        exc, tb = err_tb
                        f.write(f"{tb}\n{exc}")

            case Run.State.COMPLETE:
                for _file in (*_flags, self.df_path):
                    _try_delete_if_exists(_file)

                self.complete_flag.touch()

                if df is not None:
                    df.to_parquet(self.df_path)
            case _:
                raise ValueError(f"Unknown state {state}")


    class State(str, Enum):
        """The state of a problem."""

        PENDING = "PENDING"
        QUEUED = "QUEUED"
        RUNNING = "RUNNING"
        CRASHED = "CRASHED"
        COMPLETE = "COMPLETE"

        @classmethod
        def collect(
            cls,
            state: str | Run.State | bool | Iterable[Run.State | str],
        ) -> list[Run.State]:
            """Collect state requested."""
            match state:
                case True:
                    return list(cls)
                case False:
                    return []
                case Run.State():
                    return [state]
                case str():
                    return [cls(state)]
                case _:
                    return [cls(s) if isinstance(s, str) else s for s in state]


    @dataclass
    class Report:
        """The report of a Run."""

        run: Run
        results: list[Result]

        problem: Problem = field(init=False)

        def __post_init__(self) -> None:
            self.problem = self.run.problem

        def df(  # noqa: C901, PLR0912, PLR0915
            self,
            *,
            incumbent_trajectory: bool = False,
        ) -> pd.DataFrame:
            """Return the history as a pandas DataFrame.

            Args:
                incumbent_trajectory: Whether to only include the incumbents trajectory.

            Returns:
                The history as a pandas DataFrame.
            """
            problem = self.problem

            def _encode_result(_r: Result) -> dict[str, Any]:  # noqa: C901, PLR0912
                _rparts: dict[str, Any] = {
                    "config.id": _r.config.config_id,
                    "query.id": _r.query.query_id,
                    "result.budget_cost": _r.budget_cost,
                    "result.budget_used_total": _r.budget_used_total,
                }
                match _r.query.fidelity:
                    case None:
                        _rparts["query.fidelity.count"] = 0
                    case (name, val):
                        _rparts["query.fidelity.count"] = 1
                        _rparts["query.fidelity.1.name"] = name
                        _rparts["query.fidelity.1.value"] = val
                    case Mapping():
                        _rparts["query.fidelity.count"] = len(_r.query.fidelity)
                        for i, (k, v) in enumerate(_r.query.fidelity.items(), start=1):
                            _rparts[f"query.fidelity.{i}.name"] = k
                            _rparts[f"query.fidelity.{i}.value"] = v

                match problem.objectives:
                    case (_name, _measure):
                        _rparts["result.objective.1.value"] = _r.values[_name]
                    case Mapping():
                        for i, name in enumerate(problem.objectives, start=1):
                            _rparts[f"result.objective.{i}.value"] = _r.values[name]

                match problem.fidelities:
                    case None:
                        pass
                    case (name, _):
                        assert isinstance(_r.fidelity, tuple)
                        _rparts["result.fidelity.1.value"] = _r.fidelity[1]
                    case Mapping():
                        assert isinstance(_r.fidelity, Mapping)
                        for i, name in enumerate(problem.fidelities, start=1):
                            _rparts[f"result.fidelity.{i}.value"] = _r.fidelity[name]

                match problem.costs:
                    case None:
                        pass
                    case (name, _):
                        _rparts["result.cost.1.value"] = _r.values[name]
                    case Mapping():
                        for i, name in enumerate(problem.costs, start=1):
                            _rparts[f"result.fidelity.{i}.value"] = _r.values[name]

                _rparts["result.continuations_cost.1"] = _r.continuations_cost

                return _rparts

            parts = {}
            parts["run.name"] = self.run.name
            parts["problem.name"] = problem.name

            match problem.objectives:
                case (name, measure):
                    parts["problem.objective.count"] = 1
                    parts["problem.objective.1.name"] = name
                    parts["problem.objective.1.minimize"] = measure.minimize
                    parts["problem.objective.1.min"] = measure.bounds[0]
                    parts["problem.objective.1.max"] = measure.bounds[1]
                case Mapping():
                    list(problem.objectives)
                    parts["problem.objective.count"] = len(problem.objectives)
                    for i, (k, v) in enumerate(problem.objectives.items(), start=1):
                        parts[f"problem.objective.{i}.name"] = k
                        parts[f"problem.objective.{i}.minimize"] = v.minimize
                        parts[f"problem.objective.{i}.min"] = v.bounds[0]
                        parts[f"problem.objective.{i}.max"] = v.bounds[1]
                case _:
                    raise TypeError("Objective must be a tuple (name, measure) or a mapping")

            match problem.fidelities:
                case None:
                    parts["problem.fidelity.count"] = 0
                case (name, fid):
                    parts["problem.fidelity.count"] = 1
                    parts["problem.fidelity.1.name"] = name
                    parts["problem.fidelity.1.min"] = fid.min
                    parts["problem.fidelity.1.max"] = fid.max
                case Mapping():
                    list(problem.fidelities)
                    parts["problem.fidelity.count"] = len(problem.fidelities)
                    for i, (k, v) in enumerate(problem.fidelities.items(), start=1):
                        parts[f"problem.fidelity.{i}.name"] = k
                        parts[f"problem.fidelity.{i}.min"] = v.min
                        parts[f"problem.fidelity.{i}.max"] = v.max
                case _:
                    raise TypeError("Must be a tuple (name, fidelitiy) or a mapping")

            match self.run.benchmark.fidelities:
                case None:
                    parts["benchmark.fidelity.count"] = 0
                case (name, fid):
                    parts["benchmark.fidelity.count"] = 1
                    parts["benchmark.fidelity.1.name"] = name
                    parts["benchmark.fidelity.1.min"] = fid.min
                    parts["benchmark.fidelity.1.max"] = fid.max
                case Mapping():
                    list(self.run.benchmark.fidelities)
                    parts["benchmark.fidelity.count"] = len(self.run.benchmark.fidelities)
                    for i, (k, v) in enumerate(self.run.benchmark.fidelities.items(), start=1):
                        parts[f"benchmark.fidelity.{i}.name"] = k
                        parts[f"benchmark.fidelity.{i}.min"] = v.min
                        parts[f"benchmark.fidelity.{i}.max"] = v.max
                case _:
                    raise TypeError("Must be a tuple (name, fidelitiy) or a mapping")

            match problem.costs:
                case None:
                    parts["problem.cost.count"] = 0
                case (name, cost):
                    parts["problem.cost.count"] = 1
                    parts["problem.cost.1.name"] = name
                    parts["problem.cost.1.minimize"] = cost.minimize
                    parts["problem.cost.1.min"] = cost.bounds[0]
                    parts["problem.cost.1.max"] = cost.bounds[1]
                case Mapping():
                    list(problem.costs)
                    parts["problem.cost.count"] = len(problem.costs)
                    for i, (k, v) in enumerate(problem.costs.items(), start=1):
                        parts[f"problem.cost.{i}.name"] = k
                        parts[f"problem.cost.{i}.minimize"] = v.minimize
                        parts[f"problem.cost.{i}.min"] = v.bounds[0]
                        parts[f"problem.cost.{i}.max"] = v.bounds[1]

            _df = pd.DataFrame.from_records([_encode_result(r) for r in self.results])
            _df = _df.sort_values("result.budget_used_total", ascending=True)
            for k, v in parts.items():
                _df[k] = v

            _df["benchmark.name"] = self.run.benchmark.name
            match problem.budget:
                case TrialBudget(total):
                    _df["problem.budget.kind"] = "TrialBudget"
                    _df["problem.budget.total"] = total
                case CostBudget(total):
                    _df["problem.budget.kind"] = "CostBudget"
                    _df["problem.budget.total"] = total
                case _:
                    raise NotImplementedError(f"Unknown budget type {problem.budget}")

            _df["run.seed"] = self.run.seed
            _df["optimizer.name"] = self.run.optimizer.name

            if len(self.run.optimizer_hyperparameters) > 0:
                for k, v in self.run.optimizer_hyperparameters.items():
                    _df[f"optimizer.hp.{k}"] = v

                _df["optimizer.hp_str"] = ",".join(
                    f"{k}={v}" for k, v in self.run.optimizer_hyperparameters.items()
                )
            else:
                _df["optimizer.hp_str"] = "default"

            _df = _df.sort_values("result.budget_used_total", ascending=True)

            _df = reduce_dtypes(
                _df,
                reduce_int=True,
                reduce_float=True,
                categories=True,
                categories_exclude=("config.id", "query.id"),
            )

            if incumbent_trajectory:
                if not isinstance(self.problem.objectives, tuple):
                    raise ValueError(
                        "Incumbent trajectory only supported for single objective."
                        f" Problem {self.problem.name} has "
                        f"{len(self.problem.objectives)} objectives"
                        f" for run {self.run.name}"
                    )

                if self.problem.objectives[1].minimize:
                    _df["_tmp_"] = _df["result.objective.1.value"].cummin()
                else:
                    _df["_tmp_"] = _df["result.objective.1.value"].cummax()

                _df = _df.drop_duplicates(subset="_tmp_", keep="first").drop(columns="_tmp_")  # type: ignore

            match self.problem.objectives:
                case (_, measure):
                    _low, _high = measure.bounds
                    if not np.isinf(_low) and not np.isinf(_high):
                        _df["result.objective.1.normalized_value"] = (
                            _df["result.objective.1.value"] - _low
                        ) / (_high - _low)
                case Mapping():
                    for i, (_, measure) in enumerate(self.problem.objectives.items(), start=1):
                        _low, _high = measure.bounds
                        if not np.isinf(_low) and not np.isinf(_high):
                            _df[f"result.objective.{i}.normalized_value"] = (
                                _df[f"result.objective.{i}.value"] - _low
                            ) / (_high - _low)
                case _:
                    raise TypeError("Objective must be a tuple (name, measure) or a mapping")

            return _df

        @classmethod
        def from_df(cls, df: pd.DataFrame, run: Run) -> Run.Report:  # noqa: C901, PLR0915
            """Load a GLUEReport from a pandas DataFrame.

            Args:
                df: The dataframe to load from. Will subselect rows
                    that match the problem name.
                run: The run definition.
            """
            problem = run.problem

            def _row_to_result(series: pd.Series) -> Result:  # noqa: C901, PLR0912, PLR0915
                _row = series.to_dict()
                _result_values: dict[str, Any] = {}
                match problem.objectives:
                    case (name, _):
                        assert int(_row["problem.objective.count"]) == 1
                        assert str(_row["problem.objective.1.name"]) == name
                        _result_values[name] = _row["result.objective.1.value"]
                    case Mapping():
                        assert int(_row["problem.objective.count"]) == len(problem.objectives)
                        for i, k in enumerate(problem.objectives, start=1):
                            assert str(_row[f"problem.objective.{i}.name"]) == k
                            _result_values[k] = _row[f"result.objective.{i}.value"]
                    case _:
                        raise TypeError("Objective must be a tuple (name, measure) or a mapping")

                match problem.fidelities:
                    case None:
                        assert int(_row["problem.fidelity.count"]) == 0
                        _result_fidelity = None
                    case (name, fid):
                        assert int(_row["problem.fidelity.count"]) == 1
                        assert str(_row["problem.fidelity.1.name"]) == name
                        _result_fidelity = (name, fid.kind(_row["result.fidelity.1.value"]))
                    case Mapping():
                        assert int(_row["problem.fidelity.count"]) == len(problem.fidelities)
                        _result_fidelity = {}
                        for i, (name, fid) in enumerate(problem.fidelities.items(), start=1):
                            assert str(_row[f"problem.fidelity.{i}.name"]) == name
                            _result_fidelity[name] = fid.kind(_row[f"result.fidelity.{i}.value"])
                    case _:
                        raise TypeError("Must be a tuple (name, fidelitiy) or a mapping")

                match problem.costs:
                    case None:
                        assert int(_row["problem.cost.count"]) == 0
                    case (name, _):
                        assert int(_row["problem.cost.count"]) == 1
                        assert str(_row["problem.cost.1.name"]) == name
                        _result_values[name] = _row["result.cost.1.value"]
                    case Mapping():
                        assert int(_row["problem.cost.count"]) == len(problem.costs)
                        for i, (name, _) in enumerate(problem.costs.items(), start=1):
                            assert str(_row[f"problem.cost.{i}.name"]) == name
                            _result_values[name] = _row[f"result.cost.{i}.value"]
                    case _:
                        raise TypeError("Must be a tuple (name, fidelitiy) or a mapping")

                _query_f_count = int(_row["query.fidelity.count"])
                _query_fidelity: None | tuple[str, int | float] | dict[str, int | float]
                match _query_f_count:
                    case 0:
                        _query_fidelity = None
                    case 1:
                        _name = str(_row["query.fidelity.1.name"])
                        assert run.benchmark.fidelities is not None
                        _fid = run.benchmark.fidelities[_name]
                        _val = _fid.kind(_row["query.fidelity.1.value"])
                        _query_fidelity = (_name, _val)
                    case _:
                        _query_fidelity = {}
                        for i in range(1, _query_f_count + 1):
                            _name = str(_row[f"query.fidelity.{i}.name"])
                            assert run.benchmark.fidelities is not None
                            _fid = run.benchmark.fidelities[_name]
                            _query_fidelity[_name] = _fid.kind(_row[f"query.fidelity.{i}.value"])

                return Result(
                    query=Query(
                        config=Config(config_id=str(_row["config.id"]), values=None),
                        optimizer_info=None,
                        request_trajectory=False,
                        fidelity=_query_fidelity,
                    ),
                    budget_cost=float(_row["result.budget_cost"]),
                    budget_used_total=float(_row["result.budget_used_total"]),
                    values=_result_values,
                    fidelity=_result_fidelity,
                    trajectory=None,
                )

            this_run = df["run.name"] == run.name
            run_columns = [c for c in df.columns if c.startswith("run.")]
            problem_columns = [c for c in df.columns if c.startswith("problem.")]
            dup_rows = df[this_run].drop_duplicates(subset=run_columns + problem_columns)
            if len(dup_rows) > 1:
                raise ValueError(
                    f"Multiple run rows found for the provided df for run '{run.name}'"
                    f"\n{dup_rows}"
                )

            df = df[this_run].sort_values("result.budget_used_total")  # noqa: PD901
            return cls(
                run=run,
                results=[_row_to_result(row) for _, row in df[this_run].iterrows()],
            )

        def save(self, path: Path) -> None:
            """Save the report to a path."""
            self.df().to_parquet(path, index=False)

        @classmethod
        def from_path(cls, path: Path, problem: Run) -> Run.Report:
            """Load a report from a path."""
            df = pd.read_parquet(path)  # noqa: PD901
            return cls.from_df(df, run=problem)



from __future__ import annotations

import logging
from argparse import ArgumentParser
from pathlib import Path

import numpy as np
import pandas as pd
from hpoglue import BenchmarkDescription, Config, Query
from hpoglue.fidelity import ContinuousFidelity, ListFidelity, RangeFidelity
from scipy.stats import qmc

from hposuite.benchmarks import BENCHMARKS
from hposuite.constants import DATA_DIR

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)

def _create_sobol_grid(n_samples, dim, seed=0) -> np.ndarray:
    sobol = qmc.Sobol(d=dim, seed=seed)
    return sobol.random(n_samples)


def _scale_to_bounds(data: np.ndarray, bounds: np.ndarray) -> np.ndarray:
    return qmc.scale(data, l_bounds=bounds[0], u_bounds=bounds[1])

def _save_tabular(
    df: pd.DataFrame,
    save_dir: Path,
    name: str,
) -> None:
    """Saves the tabular data to a parquet file."""
    save_dir = save_dir / f"{name}.parquet"
    df.to_parquet(save_dir)
    logger.info(f"Saved tabular data to {save_dir.resolve()}.")


def create_tabular(
    n_samples: int,
    benchmark_desc: BenchmarkDescription,
    tabular_suite_name: str,
    task: str,
    seed: int = 0,
    data_dir: Path = DATA_DIR,
    fidelity: str | None = None,
) -> None:
    """Creates a Tabular Benchmark from the given Synthetic/Functional Benchmark."""
    space = benchmark_desc.config_space
    bench = benchmark_desc.load(benchmark_desc)
    dim = len(space.values())

    fidelity_space: tuple[str, list[int | float]] | None = None

    if fidelity:
        assert fidelity in benchmark_desc.fidelities, (
            f"Fidelity {fidelity} not found in benchmark fidelities."
        )
        match benchmark_desc.fidelities[fidelity]:
            case ContinuousFidelity():
                sampled_fids = _create_sobol_grid(n_samples, 1, seed)
                fidelity_space = (
                    fidelity,
                    _scale_to_bounds(
                        sampled_fids,
                        np.array([
                            [
                                benchmark_desc.fidelities[fidelity].min,
                                benchmark_desc.fidelities[fidelity].max
                            ]
                        ])
                    )
                )
            case ListFidelity():
                fidelity_space = (fidelity, benchmark_desc.fidelities[fidelity].values)
            case RangeFidelity():
                _min, _max, _step = (
                    benchmark_desc.fidelities[fidelity].min,
                    benchmark_desc.fidelities[fidelity].max,
                    benchmark_desc.fidelities[fidelity].stepsize
                )
                fidelity_space = (
                    fidelity,
                    np.arange(_min, _max + _step, _step)
                )
            case _:
                raise TypeError(
                    f"Unsupported fidelity type: {type(benchmark_desc.fidelities[fidelity])}"
                )

    logger.info(f"Generating {n_samples} configs using Sobol sequence for {benchmark_desc.name}.")
    sampled_configs = _create_sobol_grid(n_samples, dim, seed)
    config_dict = {}
    bounds = np.array(
        [
            np.array([hp.lower, hp.upper])
            for hp in space.values()
        ]
    ).T
    assert bounds.shape == (2, dim), "Bounds shape mismatch."
    assert bounds.shape[0] == 2, "First dimension of bounds must be 2." # noqa: PLR2004
    assert bounds.shape[1] == dim, "Dimension mismatch."
    scaled_configs = _scale_to_bounds(sampled_configs, bounds)

    logger.info(f"Querying given benchmark for {n_samples} configs.")
    for config_id, config in enumerate(scaled_configs, start=1):
        configs= {
            hp.name: config[i]
            for i, hp in enumerate(space.values())
        }
        if fidelity:
            for fid in fidelity_space[1]:
                query = Query(
                    config=Config(config_id=config_id, values=configs),
                    fidelity=(fidelity, fid)
                )
                results = bench.query(query).values
                config_dict[(config_id, fid)] = {
                    **configs,
                    **results
                }
            continue

        query = Query(
            config=Config(config_id=config_id, values=configs),
            fidelity=None
        )
        results = bench.query(query).values
        config_dict[config_id] = {
            **configs,
            **results
        }

    logger.info(f"Saving tabular data for {benchmark_desc.name}.")
    _df = pd.DataFrame.from_dict(config_dict, orient="index")
    if fidelity:
        _df.index.names = ["config_id", fidelity]
    else:
        _df.index.name = "config_id"
    print(_df.head())  # noqa: T201
    save_dir = data_dir / tabular_suite_name
    save_dir.mkdir(parents=True, exist_ok=True)
    _save_tabular(_df, save_dir, f"{task}")




if __name__ == "__main__":
    parser = ArgumentParser()
    parser.add_argument(
        "--n_samples", "-n",
        type=int,
        default=2000,
        help="Number of configs to generate."
    )
    parser.add_argument(
        "--benchmark", "-b",
        type=str,
        required=True,
        help="Benchmark to create tabular benchmark for."
    )
    parser.add_argument(
        "--seed", "-s",
        type=int,
        default=0,
        help="Seed for reproducibility."
    )
    parser.add_argument(
        "--data_dir", "-d",
        type=str,
        default=DATA_DIR,
        help="Directory to save the tabular data."
    )
    parser.add_argument(
        "--tabular_suite_name", "-suite",
        type=str,
        required=True,
        help="Name of the tabular benchmark suite. Eg: bbob_tabular"
    )
    parser.add_argument(
        "--task", "-t",
        type=str,
        required=True,
        help="Name of the tabular benchmark. Eg: bbob function name: f1-2-0"
    )
    parser.add_argument(
        "--fidelity", "-f",
        type=str,
        default=None,
        help="Fidelity key to include in the tabular benchmark."
    )
    args = parser.parse_args()

    benchmark = BENCHMARKS[args.benchmark]

    if not isinstance(benchmark, BenchmarkDescription):
        benchmark = benchmark.desc

    if not isinstance(args.data_dir, Path):
        args.data_dir = Path(args.data_dir)

    create_tabular(
        n_samples=args.n_samples,
        benchmark_desc=benchmark,
        tabular_suite_name=args.tabular_suite_name,
        task=args.task,
        data_dir=args.data_dir,
        seed=args.seed,
        fidelity=args.fidelity
    )




from __future__ import annotations


class OptBenchNotInstalledError(ImportError):
    """Raised when a benchmark or optimizer is not installed."""
    def __init__(self, module: str, msg: str) -> None:
        """Initialize the exception."""
        import re
        not_installed_str = re.search(r"['\"]([^'\"]+)['\"]", msg)
        not_installed_str = not_installed_str.group(1) if not_installed_str else "Unknown"
        match not_installed_str:
            case "Unknown":
                super()._init_(msg)
            case _:
                super().__init__(
                    f"{not_installed_str} is not installed in module {module}. "
                    "Please install it first."
                )
        self.msg = msg
        self.name = not_installed_str
        self.module = module

    def __str__(self) -> str:
        if self.name == "Unknown":
            return self.msg
        return (
            f"{self.name} is not installed in module {self.module}. Please install it first."
        )

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.name})"


from __future__ import annotations

import argparse
import logging

from hposuite.optimizers import (
    BB_OPTIMIZERS,
    MF_OPTIMIZERS,
    MO_OPTIMIZERS,
    OPTIMIZERS,
    SO_OPTIMIZERS,
)

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

def main(
    list: str
):
    match list:
        case "all":
            logger.info(OPTIMIZERS.keys())
        case "mf":
            logger.info(MF_OPTIMIZERS.keys())
        case "nonmf":
            logger.info(BB_OPTIMIZERS.keys())
        case "mo":
            logger.info(MO_OPTIMIZERS.keys())
        case "so":
            logger.info(SO_OPTIMIZERS.keys())


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--list",
        required=True,
        choices=["all", "mf", "nonmf", "mo", "so"],
        help="List all Optimizers or filter by type"
        "\nall: list all Optimizers"
        "\nmf: list multi-fidelity Optimizers"
        "\nnonmf: list non-multi-fidelity Optimizers"
        "\nmo: list multi-objective Optimizers"
        "\nso: list single-objective Optimizers",
    )
    args = parser.parse_args()
    main(args.list)



from __future__ import annotations

# https://syne-tune.readthedocs.io/en/latest/examples.html#ask-tell-interface
# https://syne-tune.readthedocs.io/en/latest/examples.html#ask-tell-interface-for-hyperband
import datetime
from abc import abstractmethod
from collections.abc import Mapping
from pathlib import Path
from typing import TYPE_CHECKING, Any, Literal

import ConfigSpace as CS  # noqa: N817
from hpoglue import Config, Optimizer, Problem, Query
from hpoglue.env import Env

if TYPE_CHECKING:
    from hpoglue import Result
    from syne_tune.config_space import (
        Domain,
    )
    from syne_tune.optimizer.scheduler import TrialScheduler


class SyneTuneOptimizer(Optimizer):
    """Base class for SyneTune Optimizers."""

    name = "SyneTune_base"

    env = Env(
        name="syne_tune-0.13.0",
        python_version="3.10.12",
        requirements=("syne_tune==0.13.0",),
    )

    @abstractmethod
    def __init__(
        self,
        *,
        problem: Problem,
        seed: int,
        working_directory: Path,
        optimizer: TrialScheduler,
    ):
        """Create a SyneTune Optimizer instance for a given problem statement."""
        working_directory.mkdir(parents=True, exist_ok=True)
        self.problem = problem
        self.optimizer = optimizer
        self._counter = 0

    def ask(self) -> Query:
        """Get a configuration from the optimizer."""
        from syne_tune.backend.trial_status import Trial

        self._counter += 1
        trial_suggestion = self.optimizer.suggest(self._counter)
        assert trial_suggestion is not None
        assert trial_suggestion.config is not None
        name = str(self._counter)
        trial = Trial(
            trial_id=self._counter,
            config=trial_suggestion.config,
            creation_time=datetime.datetime.now(),
        )

        fidelity = None
        if isinstance(self.problem.fidelities, tuple):
            fidelity = trial_suggestion.config.pop(self.problem.fidelities[0])

        # TODO: How to get the fidelity??
        return Query(
            config=Config(config_id=name, values=trial.config),
            fidelity=fidelity,
            optimizer_info=trial,
        )

    def tell(self, result: Result) -> None:
        """Update the SyneTune Optimizer with the result of a query."""
        match self.problem.objectives:
            case Mapping():
                results_obj_dict = {
                    key: result.values[key]
                    for key in result.values
                    if key in self.problem.objectives
                }
            case (metric_name, _):
                results_obj_dict = {metric_name: result.values[metric_name]}
            case _:
                raise TypeError("Objective must be a string or a list of strings")

        self.optimizer.on_trial_complete(
            trial=result.query.optimizer_info,  # type: ignore
            result=results_obj_dict,
        )


class SyneTuneBO(SyneTuneOptimizer):
    """SyneTune Bayesian Optimization."""

    name = "SyneTune_BO"
    support = Problem.Support(
        fidelities=(None,),
        objectives=("single",),
        cost_awareness=(None,),
        tabular=False,
    )

    mem_req_mb = 1024

    def __init__(  # noqa: C901, PLR0912
        self,
        *,
        problem: Problem,
        seed: int,
        working_directory: Path,
        **kwargs: Any,
    ):
        """Create a SyneTune Bayesian Optimization instance for a given problem statement.

        Args:
            problem: The problem statement.
            seed: The random seed.
            working_directory: The working directory to store the results.
            **kwargs: Additional arguments for the BayesianOptimization.
        """
        from syne_tune.optimizer.baselines import BayesianOptimization

        match problem.fidelities:
            case None:
                pass
            case tuple():
                clsname = self.__class__.__name__
                # raise ValueError(f"{clsname} does not multi-fidelity spaces")
            case Mapping():
                clsname = self.__class__.__name__
                raise NotImplementedError(f"{clsname} does not support many-fidelity")
            case _:
                raise TypeError("fidelity_space must be a list, dict or None")

        match problem.costs:
            case None:
                pass
            case tuple() | Mapping():
                clsname = self.__class__.__name__
                raise ValueError(f"{clsname} does not support cost-awareness")
            case _:
                raise TypeError("cost_awareness must be a list, dict or None")

        metric_name: str
        mode: Literal["min", "max"]
        match problem.objectives:
            case (name, metric):
                metric_name = name
                mode = "min" if metric.minimize else "max"
            case Mapping():
                raise NotImplementedError(
                    "# TODO: Multiobjective not yet implemented for SyneTuneBO!"
                )
            case _:
                raise TypeError("Objective must be a string or a list of strings")

        synetune_cs: dict[str, Domain]

        config_space = problem.config_space
        match config_space:
            case CS.ConfigurationSpace():
                synetune_cs = configspace_to_synetune_configspace(config_space)
            case list():
                raise ValueError("SyneTuneBO does not support tabular benchmarks")
            case _:
                raise TypeError("config_space must be of type ConfigSpace.ConfigurationSpace")

        if isinstance(problem.fidelities, tuple):
            synetune_cs[problem.fidelities[0]] = problem.fidelities[1]  # TODO: Check this

        super().__init__(
            problem=problem,
            seed=seed,
            working_directory=working_directory,
            optimizer=BayesianOptimization(
                config_space=synetune_cs,
                metric=metric_name,
                mode=mode,
                random_seed=seed,
                **kwargs,
            ),
        )


class SyneTuneBOHB(SyneTuneOptimizer):
    """SyneTune BOHB."""

    name = "SyneTune_BOHB"
    support = Problem.Support(
        fidelities=(None, "single"),
        objectives=("single",),
        cost_awareness=(None,),
        tabular=False,
    )

    mem_req_mb = 1024

    def __init__(  # noqa: C901, PLR0912
        self,
        *,
        problem: Problem,
        seed: int,
        working_directory: Path,
        **kwargs: Any,
    ):
        """Create a SyneTune BOHB instance for a given problem statement.

        Args:
            problem: The problem statement.
            seed: The random seed.
            working_directory: The working directory to store the results.
            **kwargs: Additional arguments for the BayesianOptimization.
        """
        from syne_tune.optimizer.baselines import BOHB

        match problem.fidelities:
            case None:
                min_fidelity = None
                max_fidelity = None
            case (_, fidelity):
                min_fidelity = fidelity.min
                max_fidelity = fidelity.max
            case Mapping():
                clsname = self.__class__.__name__
                raise NotImplementedError(f"{clsname} does not support many-fidelity")
            case _:
                raise TypeError("fidelity_space must be a list, dict or None")

        match problem.costs:
            case None:
                pass
            case tuple() | Mapping():
                clsname = self.__class__.__name__
                raise ValueError(f"{clsname} does not support cost-awareness")
            case _:
                raise TypeError("cost_awareness must be a list, dict or None")

        metric_name: str
        mode: Literal["min", "max"]
        match problem.objectives:
            case (name, metric):
                metric_name = name
                mode = "min" if metric.minimize else "max"
            case Mapping():
                raise NotImplementedError(
                    "# TODO: Multiobjective not yet implemented for SyneTuneBO!"
                )
            case _:
                raise TypeError("Objective must be a string or a list of strings")

        synetune_cs: dict[str, Domain]

        config_space = problem.config_space
        match config_space:
            case CS.ConfigurationSpace():
                synetune_cs = configspace_to_synetune_configspace(config_space)
            case list():
                raise ValueError("SyneTuneBO does not support tabular benchmarks")
            case _:
                raise TypeError("config_space must be of type ConfigSpace.ConfigurationSpace")

        if isinstance(problem.fidelities, tuple):
            synetune_cs[problem.fidelities[0]] = problem.fidelities[1]  # TODO: Check this

        super().__init__(
            problem=problem,
            seed=seed,
            working_directory=working_directory,
            optimizer=BOHB(
            config_space=self.synetune_cs,
            mode = mode,
            metric = metric_name,
            random_seed = seed,
            type="stopping",
            max_t = problem.fidelities[1].max,
            resource_attr = problem.fidelities[0],
            grace_period=1,
            eta=3,
            )
        )


def configspace_to_synetune_configspace(  # noqa: C901
    config_space: CS.ConfigurationSpace,
) -> dict[str, Domain | Any]:
    """Convert ConfigSpace to SyneTune config_space."""
    from syne_tune.config_space import (
        choice,
        lograndint,
        loguniform,
        ordinal,
        randint,
        uniform,
    )

    if any(config_space.get_conditions()):
        raise NotImplementedError("ConfigSpace with conditions not supported")

    if any(config_space.get_forbiddens()):
        raise NotImplementedError("ConfigSpace with forbiddens not supported")

    synetune_cs: dict[str, Domain | Any] = {}
    for hp in config_space.get_hyperparameters():
        match hp:
            case CS.OrdinalHyperparameter():
                synetune_cs[hp.name] = ordinal(hp.sequence)
            case CS.CategoricalHyperparameter() if hp.weights is not None:
                raise NotImplementedError("CategoricalHyperparameter with weights not supported")
            case CS.CategoricalHyperparameter():
                synetune_cs[hp.name] = choice(hp.choices)
            case CS.UniformIntegerHyperparameter() if hp.log:
                synetune_cs[hp.name] = lograndint(hp.lower, hp.upper)
            case CS.UniformIntegerHyperparameter():
                synetune_cs[hp.name] = randint(hp.lower, hp.upper)
            case CS.UniformFloatHyperparameter() if hp.log:
                synetune_cs[hp.name] = loguniform(hp.lower, hp.upper)
            case CS.UniformFloatHyperparameter():
                synetune_cs[hp.name] = uniform(hp.lower, hp.upper)
            case CS.Constant():
                synetune_cs[hp.name] = hp.value
            case _:
                raise ValueError(f"Hyperparameter {hp.name} of type {type(hp)} is not supported")

    return synetune_cs



from __future__ import annotations

from collections.abc import Mapping
from pathlib import Path
from typing import TYPE_CHECKING, Any

import numpy as np
from ConfigSpace import ConfigurationSpace
from hebo.design_space.design_space import DesignSpace
from hebo.optimizers.hebo import HEBO
from hpoglue import Config, Optimizer, Problem, Query
from hpoglue.env import Env

if TYPE_CHECKING:
    from hpoglue import Result

class HEBOOptimizer(Optimizer):
    """The HEBO Optimizer."""

    name = "HEBO"

    env = Env(
        name="hebo-0.3.6",
        python_version="3.10",
        requirements=("hebo==0.3.6",),
    )

    support = Problem.Support(
        fidelities=(None,),
        objectives=("single", "many"),
        cost_awareness=(None,),
        tabular=False,
        priors=False,
    )

    mem_req_mb = 1024


    def __init__(
        self,
        *,
        problem: Problem,
        seed: int,
        working_directory: Path,  # noqa: ARG002
        **kwargs: Any,
    ) -> None:
        """Initialize the HEBO optimizer."""
        self.problem = problem
        self.config_space: ConfigurationSpace = problem.config_space
        self._space: DesignSpace
        match self.config_space:
            case ConfigurationSpace():
                self._space = _configspace_to_hebo_designspace(self.config_space)
            case list():
                raise ValueError("HEBO does not support list-type config spaces!")
            case _:
                raise TypeError("Config space must be a list or a ConfigurationSpace!")

        self.optimizer: HEBO = HEBO(
            space=self._space,
            scramble_seed=seed,
            **kwargs
        )
        self._trial_counter = 0


    def ask(self) -> Query:
        """Asks the optimizer to suggest a new configuration to evaluate.

        Returns:
            Query: A new query containing the suggested configuration, fidelity, and
                    other relevant optimizer information.
        """
        match self.problem.fidelities:
            case None:
                suggestion = self.optimizer.suggest(1)
                hyp = suggestion.iloc[0].to_dict()

                from ConfigSpace.hyperparameters import OrdinalHyperparameter
                for k in hyp:
                    hp_type = self._space.paras[k]
                    if hp_type.is_numeric and hp_type.is_discrete and not np.isnan(hyp[k]):
                        hyp[k] = int(hyp[k])
                        # Now we need to check if it is an ordinal hp
                        hp_k = self.config_space[k]
                        if isinstance(hp_k, OrdinalHyperparameter):
                            hyp[k] = hp_k.sequence[hyp[k]]
                return Query(
                    config=Config(config_id=self._trial_counter, values=hyp),
                    fidelity=None,
                    optimizer_info=suggestion,
                )
            case tuple():
                raise ValueError("HEBO does not support multi-fidelity optimization!")
            case Mapping():
                raise ValueError("HEBO does not support many-fidelity optimization!")
            case _:
                raise TypeError("Fidelity must be None, a tuple or a Mapping!")


    def tell(self, result: Result) -> None:
        """Updates the optimizer with the result of an evaluation.

        Parameters:
        result: The result of an evaluation containing the objective values
                and other relevant information.
        """
        match self.problem.objectives:
            case (name, _):
                _values = np.asarray([result.values[name]])
            case Mapping():
                raise ValueError("# TODO: Multiobjective not yet implemented for HEBO")
            case _:
                raise TypeError("Objective must be a string or a list of strings!")

        match self.problem.costs:
            case None:
                pass
            case tuple():
                raise ValueError("Cost-aware optimization not supported by HEBO!")
            case Mapping():
                raise ValueError("Cost-aware optimization not supported by HEBO!")
            case _:
                raise TypeError("Cost must be None or a Mapping!")

        self.optimizer.observe(
            result.query.optimizer_info,
            _values,
        )


def _configspace_to_hebo_designspace(config_space: ConfigurationSpace) -> DesignSpace:  # noqa: C901, PLR0912
    """Convert ConfigSpace to HEBO design space."""
    from ConfigSpace.hyperparameters import (
        CategoricalHyperparameter,
        Constant,
        FloatHyperparameter,
        IntegerHyperparameter,
        OrdinalHyperparameter,
    )
    if len(config_space.conditions) > 0:
        raise NotImplementedError("Conditions are not yet supported!")

    if len(config_space.forbidden_clauses) > 0:
        raise NotImplementedError("Forbiddens are not yet supported!")

    hebo_space: list[dict[str, Any]] = []

    for hp in list(config_space.values()):
        if isinstance(hp, IntegerHyperparameter):
            if hp.log:
                hebo_space.append(
                    {"name": hp.name, "type": "pow_int", "lb": hp.lower, "ub": hp.upper}
                )
            else:
                hebo_space.append({"name": hp.name, "type": "int", "lb": hp.lower, "ub": hp.upper})
        elif isinstance(hp, FloatHyperparameter):
            if hp.log:
                hebo_space.append({"name": hp.name, "type": "pow", "lb": hp.lower, "ub": hp.upper})
            else:
                hebo_space.append({"name": hp.name, "type": "num", "lb": hp.lower, "ub": hp.upper})
        elif isinstance(hp, CategoricalHyperparameter):
            hebo_space.append({"name": hp.name, "type": "cat", "categories": hp.choices})
        elif isinstance(hp, OrdinalHyperparameter):
            hebo_space.append({
                "name": hp.name,
                "type": "step_int",
                "lb": 0,
                "ub": len(hp.sequence),
                "step": 1,
            })
        elif isinstance(hp, Constant):
            hebo_space.append({"name": hp.name, "type": "cat", "categories": [hp.value]})
        else:
            raise ValueError(f"Unknown hyperparameter type: {hp.__class__.__name__}")
    return DesignSpace().parse(hebo_space)


from __future__ import annotations

import importlib
import logging
from typing import TYPE_CHECKING

from hposuite.exceptions import OptBenchNotInstalledError

if TYPE_CHECKING:
    from hpoglue import Optimizer

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)

modules = [
    ("hposuite.optimizers.dehb", "DEHB_Optimizer"),
    ("hposuite.optimizers.hebo", "HEBOOptimizer"),
    ("hposuite.optimizers.nevergrad", "NevergradOptimizer"),
    ("hposuite.optimizers.optuna", "OptunaOptimizer"),
    ("hposuite.optimizers.random_search", "RandomSearch", "RandomSearchWithPriors"),
    ("hposuite.optimizers.scikit_optimize", "SkoptOptimizer"),
    ("hposuite.optimizers.smac", "SMAC_BO", "SMAC_Hyperband", "SMAC_BOHB"),
]

imported_opt_cls = []

for module_name, *attrs in modules:
    try:
        module = importlib.import_module(module_name)
        for attr in attrs:
            opt = getattr(module, attr)
            imported_opt_cls.append(opt)
    except ImportError as e:
        logger.warning(OptBenchNotInstalledError(module_name, e.msg))



OPTIMIZERS: dict[str, type[Optimizer]] = {opt.name: opt for opt in imported_opt_cls}

MF_OPTIMIZERS: dict[str, type[Optimizer]] = {}
BB_OPTIMIZERS: dict[str, type[Optimizer]] = {}
MO_OPTIMIZERS: dict[str, type[Optimizer]] = {}
SO_OPTIMIZERS: dict[str, type[Optimizer]] = {}

for name, opt in OPTIMIZERS.items():
    if "single" in opt.support.fidelities:
        MF_OPTIMIZERS[name] = opt
    else:
        BB_OPTIMIZERS[name] = opt
    if "many" in opt.support.objectives:
        MO_OPTIMIZERS[name] = opt
    if "single" in opt.support.objectives:
        SO_OPTIMIZERS[name] = opt
__all__ = [
    "OPTIMIZERS",
    "MF_OPTIMIZERS",
    "BB_OPTIMIZERS",
    "MO_OPTIMIZERS",
    "SO_OPTIMIZERS",
]



from __future__ import annotations

import argparse
import logging
from collections.abc import Mapping
from itertools import cycle
from pathlib import Path
from typing import Any, Literal, TypeAlias, TypeVar

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import yaml

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

matplotlib_logger = logging.getLogger("matplotlib")
matplotlib_logger.setLevel(logging.ERROR)

DataContainer: TypeAlias = np.ndarray | (pd.DataFrame | pd.Series)
D = TypeVar("D", bound=DataContainer)

ROOT_DIR = Path(__file__).absolute().resolve().parent.parent.parent

SEED_COL = "run.seed"
PROBLEM_COL = "problem.name"
OPTIMIZER_COL = "optimizer.name"
BENCHMARK_COL = "benchmark.name"
HP_COL = "optimizer.hp_str"
SINGLE_OBJ_NAME = "problem.objective.1.name"
SINGLE_OBJ_COL = "result.objective.1.value"
SINGLE_OBJ_MINIMIZE_COL = "problem.objective.1.minimize"
SECOND_OBJ_NAME = "problem.objective.2.name"
SECOND_OBJ_COL = "result.objective.2.value"
SECOND_OBJ_MINIMIZE_COL = "problem.objective.2.minimize"
BUDGET_USED_COL = "result.budget_used_total"
BUDGET_TOTAL_COL = "problem.budget.total"
FIDELITY_COL = "result.fidelity.1.value"
FIDELITY_NAME_COL = "problem.fidelity.1.name"
FIDELITY_MIN_COL = "problem.fidelity.1.min"
FIDELITY_MAX_COL = "problem.fidelity.1.max"
BENCHMARK_COUNT_FIDS = "benchmark.fidelity.count"
BENCHMARK_FIDELITY_NAME = "benchmark.fidelity.1.name"
BENCHMARK_FIDELITY_COL = "benchmark.fidelity.1.value"
BENCHMARK_FIDELITY_MIN_COL = "benchmark.fidelity.1.min"
BENCHMARK_FIDELITY_MAX_COL = "benchmark.fidelity.1.max"
CONTINUATIONS_COL = "result.continuations_cost.1"


def plot_results(  # noqa: C901, PLR0912, PLR0913, PLR0915
    *,
    report: dict[str, Any],
    objective: str,
    fidelity: str | None,
    cost: str | None,
    to_minimize: bool,
    save_dir: Path,
    benchmarks_name: str,
    regret_bound: float | None = None,  # noqa: ARG001
    figsize: tuple[int, int] = (20, 10),
    logscale: bool = False,
    error_bars: Literal["std", "sem"] = "std",
) -> None:
    """Plot the results for the optimizers on the given benchmark."""
    marker_list = [
        "o",
        "X",
        "^",
        "H",
        ">",
        "^",
        "p",
        "P",
        "*",
        "h",
        "<",
        "s",
        "x",
        "+",
        "D",
        "d",
        "|",
        "_",
    ]
    markers = cycle(marker_list)
    colors = plt.rcParams["axes.prop_cycle"].by_key()["color"]  # type: ignore
    colors_mean = cycle(colors)
    optimizers = list(report.keys())
    plt.figure(figsize=figsize)
    optim_res_dict = {}
    continuations = False
    for instance in optimizers:
        logger.info(f"Plotting {instance}")
        optim_res_dict[instance] = {}
        seed_cost_dict = {}
        seed_cont_dict = {}
        for seed in report[instance]:
            results = report[instance][seed]["results"]
            cost_list: pd.Series = results[SINGLE_OBJ_COL].values.astype(np.float64)

            budget_type = "TrialBudget" if fidelity is None else "FidelityBudget"
            match budget_type:
                case "FidelityBudget":
                    if FIDELITY_COL in results.columns:
                        budget_list = results[FIDELITY_COL].values.astype(np.float64)
                        budget_list = np.cumsum(budget_list)
                        budget_type = "FidelityBudget"
                    else:
                        budget_list = np.cumsum(
                            results[BENCHMARK_FIDELITY_MAX_COL].values.astype(np.float64)
                        )
                case "TrialBudget":
                    budget_list = results[BUDGET_USED_COL].values.astype(np.float64)
                case _:
                    raise NotImplementedError(f"Budget type {budget_type} not implemented")
            budget = budget_list[-1]

            if (
                CONTINUATIONS_COL in results.columns
                and
                not pd.isna(results[CONTINUATIONS_COL].iloc[0])
            ):
                continuations = True
                continuations_list = results[CONTINUATIONS_COL].values.astype(np.float64)
                continuations_list = np.cumsum(continuations_list)

            seed_cost_dict[seed] = pd.Series(cost_list, index=budget_list)
            if continuations:
                seed_cont_dict[seed] = pd.Series(cost_list, index=continuations_list)

        seed_cost_df = pd.DataFrame(seed_cost_dict)
        seed_cost_df = seed_cost_df.ffill(axis=0)
        seed_cost_df = seed_cost_df.dropna(axis=0)
        means = pd.Series(seed_cost_df.mean(axis=1), name=f"means_{instance}")
        match error_bars:
            case "std":
                error = pd.Series(seed_cost_df.std(axis=1), name=f"std_{instance}")
            case "sem":
                error = pd.Series(seed_cost_df.sem(axis=1), name=f"sem_{instance}")
            case _:
                raise ValueError(f"Unsupported error bars type {error_bars}")
        optim_res_dict[instance]["means"] = means
        optim_res_dict[instance]["error"] = error
        means = means.cummin() if to_minimize else means.cummax()
        means = means.drop_duplicates()
        error = error.loc[means.index]
        means[budget] = means.iloc[-1]
        error[budget] = error.iloc[-1]
        col_next = next(colors_mean)

        plt.step(
            means.index,
            means,
            where="post",
            label=instance,
            marker=next(markers),
            markersize=10,
            markerfacecolor="#ffffff",
            markeredgecolor=col_next,
            markeredgewidth=2,
            color=col_next,
            linewidth=3,
        )
        plt.fill_between(
            means.index,
            means - error,
            means + error,
            alpha=0.2,
            step="post",
            color=col_next,
            edgecolor=col_next,
            linewidth=2,
        )

        #For plotting continuations
        if continuations:
            seed_cont_df = pd.DataFrame(seed_cont_dict)
            seed_cont_df = seed_cont_df.ffill(axis=0)
            seed_cont_df = seed_cont_df.dropna(axis=0)
            means_cont = pd.Series(seed_cont_df.mean(axis=1), name=f"means_{instance}")
            std_cont = pd.Series(seed_cont_df.std(axis=1), name=f"std_{instance}")
            optim_res_dict[instance]["cont_means"] = means_cont
            optim_res_dict[instance]["cont_std"] = std_cont
            means_cont = means_cont.cummin() if to_minimize else means_cont.cummax()
            means_cont = means_cont.drop_duplicates()
            std_cont = std_cont.loc[means_cont.index]
            col_next = next(colors_mean)

            plt.step(
                means_cont.index,
                means_cont,
                where="post",
                label=f"{instance}_w_continuations",
                marker=next(markers),
                markersize=10,
                markerfacecolor="#ffffff",
                markeredgecolor=col_next,
                markeredgewidth=2,
                color=col_next,
                linewidth=3,
            )
            plt.fill_between(
                means_cont.index,
                means_cont - std_cont,
                means_cont + std_cont,
                alpha=0.2,
                step="post",
                color=col_next,
                edgecolor=col_next,
                linewidth=2,
            )
    plt.xlabel(f"{budget_type}")
    plt.ylabel(f"{objective}")
    plot_suffix = (
        f"{benchmarks_name}, {objective=}, \n{fidelity=}, {cost=}, "
        f"{budget_type}={budget}, {to_minimize=}, {error_bars=}"
    )
    plt.title(f"Plot for optimizers on {plot_suffix}")
    if logscale:
        plt.xscale("log")
    if len(optimizers) == 1:
        plt.title(f"Performance of {optimizers[0]} on {plot_suffix}")
    plt.legend()
    save_dir.mkdir(parents=True, exist_ok=True)

    optimizers = ",".join(optimizers)

    plot_suffix = plot_suffix.replace("\n", "")

    save_path = save_dir / f"{optimizers}.{plot_suffix}.png"
    plt.savefig(save_path)
    logger.info(f"Saved plot to {save_path.absolute()}")


def agg_data(  # noqa: C901, PLR0912, PLR0915
    study_dir: Path,
    save_dir: Path,
    figsize: tuple[int, int] = (20, 10),
    *,
    benchmark_spec: str | list[str] | None = None,
    optimizer_spec: str | list[str] | None = None,
    error_bars: Literal["std", "sem"] = "std",
    logscale: bool = False,
    budget_type: Literal["TrialBudget", "FidelityBudget", None] = None,
) -> None:
    """Aggregate the data from the run directory for plotting."""
    objective: str | None = None
    minimize = True

    with (study_dir / "study_config.yaml").open("r") as f:
        study_config = yaml.safe_load(f)

    all_benches = [(bench.pop("name"), bench) for bench in study_config["benchmarks"]]

    match benchmark_spec:
        case None:
            benchmarks_in_dir = [
                (f.name.split("benchmark=")[-1].split(".")[0])
                for f in study_dir.iterdir() if f.is_dir() and "benchmark=" in f.name]
            benchmarks_in_dir = list(set(benchmarks_in_dir))
            logger.info(f"Found benchmarks: {benchmarks_in_dir}")
        case str():
            benchmarks_in_dir = [benchmark_spec]
            logger.info(f"Benchmarks specified: {benchmarks_in_dir}")
        case list():
            benchmarks_in_dir = benchmark_spec
            logger.info(f"Benchmarks specified: {benchmarks_in_dir}")
        case _:
            raise ValueError(f"Unsupported type for benchmark_spec: {type(benchmark_spec)}")

    match optimizer_spec:
        case None:
            optimizers_in_dir = None
        case str():
            optimizers_in_dir = [optimizer_spec]
        case list():
            optimizers_in_dir = optimizer_spec
        case _:
            raise ValueError(f"Unsupported type for optimizer_spec: {type(optimizer_spec)}")

    benchmarks_dict: Mapping[str, Mapping[tuple[str, str, str], list[pd.DataFrame]]] = {}

    for benchmark in benchmarks_in_dir:
        logger.info(f"Processing benchmark: {benchmark}")
        for file in study_dir.rglob("*.parquet"):
            if benchmark not in file.name:
                continue
            if (
                optimizers_in_dir is not None
                and not any(spec in file.name for spec in optimizers_in_dir)
            ):
                continue
            _df = pd.read_parquet(file)

            benchmark_name = file.name.split("benchmark=")[-1].split(".")[0]

            with (file.parent / "run_config.yaml").open("r") as f:
                run_config = yaml.safe_load(f)
            objectives = run_config["problem"]["objectives"]
            if not isinstance(objectives, str) and len(objectives) > 1:
                raise NotImplementedError("Plotting not yet implemented for multi-objective runs.")
            fidelities = run_config["problem"]["fidelities"]
            if fidelities and not isinstance(fidelities, str) and len(fidelities) > 1:
                raise NotImplementedError("Plotting not yet implemented for many-fidelity runs.")

            # Add default benchmark fidelity to a blackbox Optimizer to compare it
            # alongside MF optimizers if the latter exist in the study
            bench_num_fids = _df[BENCHMARK_COUNT_FIDS].iloc[0]
            if fidelities is None and budget_type != "TrialBudget" and bench_num_fids >= 1:
                fid = next(
                    bench[1]["fidelities"]
                    for bench in all_benches
                    if bench[0] == benchmark_name
                )
                if fid == _df[BENCHMARK_FIDELITY_NAME].iloc[0]:
                # Study config is saved in such a way that if Blackbox Optimizers
                # are used along with MF optimizers on MF benchmarks, the "fidelities"
                # key in the benchmark instance in the study config is set to the fidelity
                # being used by the MF optimizers. In that case, there is no benchmark
                # instance with fidelity as None. In case of multiple fidelities being used
                # for the same benchmark, separate benchmark instances are created
                # for each fidelity.
                # If only Blackbox Optimizers are used in the study, there is only one
                # benchmark instance with fidelity as None.
                # When a problem with a Blackbox Optimizer is used on a MF benchmark,
                # each config is queried at the highest available 'first' fidelity in the
                # benchmark. Hence, we only set `fidelities` to `fid` if the benchmark instance
                # is the one with the default fidelity, else it would be incorrect.
                    fidelities = fid

            costs = run_config["problem"]["costs"]
            if costs:
                raise NotImplementedError(
                    "Cost-aware optimization not yet implemented in hposuite."
                )
            seed = int(run_config["seed"])
            all_plots_dict = benchmarks_dict.setdefault(benchmark, {})
            conf_tuple = (objectives, fidelities, costs)
            if conf_tuple not in all_plots_dict:
                all_plots_dict[conf_tuple] = [_df]
            else:
                all_plots_dict[conf_tuple].append(_df)


    for benchmark, conf_dict in benchmarks_dict.items():
        for conf_tuple, _all_dfs in conf_dict.items():
            df_agg = {}
            objective = conf_tuple[0]
            fidelity = conf_tuple[1]
            cost = conf_tuple[2]
            for _df in _all_dfs:
                if _df.empty:
                    continue
                instance = _df[OPTIMIZER_COL].iloc[0]
                if _df[HP_COL].iloc[0] is not None:
                    instance = f"{instance}_{_df[HP_COL].iloc[0]}"
                minimize = _df[SINGLE_OBJ_MINIMIZE_COL].iloc[0]
                seed = _df[SEED_COL].iloc[0]
                if instance not in df_agg:
                    df_agg[instance] = {}
                if int(seed) not in df_agg[instance]:
                    df_agg[instance][int(seed)] = {"results": _df}
                assert objective is not None
                benchmark_name = _df[BENCHMARK_COL].iloc[0]
            plot_results(
                report=df_agg,
                objective=objective,
                fidelity=fidelity,
                cost=cost,
                to_minimize=minimize,
                save_dir=save_dir,
                benchmarks_name=benchmark.split("benchmark=")[-1].split(".")[0],
                figsize=figsize,
                logscale=logscale,
                error_bars=error_bars,
            )
            df_agg.clear()


def scale(
    unit_xs: int | float | np.number | np.ndarray | pd.Series,
    to: tuple[int | float | np.number, int | float | np.number],
) -> float | np.number | np.ndarray | pd.Series:
    """Scale values from unit range to a new range.

    >>> scale(np.array([0.0, 0.5, 1.0]), to=(0, 10))
    array([ 0.,  5., 10.])

    Parameters
    ----------
    unit_xs:
        The values to scale

    to:
        The new range

    Returns:
    -------
        The scaled values
    """
    return unit_xs * (to[1] - to[0]) + to[0]  # type: ignore


def normalize(
    x: int | float | np.number | np.ndarray | pd.Series,
    *,
    bounds: tuple[int | float | np.number, int | float | np.number],
) -> float | np.number | np.ndarray | pd.Series:
    """Normalize values to the unit range.

    >>> normalize(np.array([0.0, 5.0, 10.0]), bounds=(0, 10))
    array([0. , 0.5, 1. ])

    Parameters
    ----------
    x:
        The values to normalize

    bounds:
        The bounds of the range

    Returns:
    -------
        The normalized values
    """
    if bounds == (0, 1):
        return x

    return (x - bounds[0]) / (bounds[1] - bounds[0])  # type: ignore


def rescale(
    x: int | float | np.number | np.ndarray | pd.Series,
    frm: tuple[int | float | np.number, int | float | np.number],
    to: tuple[int | float | np.number, int | float | np.number],
) -> float | np.ndarray | pd.Series:
    """Rescale values from one range to another.

    >>> rescale(np.array([0, 10, 20]), frm=(0, 100), to=(0, 10))
    array([0, 1, 2])

    Parameters
    ----------
    x:
        The values to rescale

    frm:
        The original range

    to:
        The new range

    Returns:
    -------
        The rescaled values
    """
    if frm != to:
        normed = normalize(x, bounds=frm)
        scaled = scale(unit_xs=normed, to=to)
    else:
        scaled = x

    match scaled:
        case int() | float() | np.number():
            return float(scaled)
        case np.ndarray() | pd.Series():
            return scaled.astype(np.float64)
        case _:
            raise ValueError(f"Unsupported type {type(x)}")


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Plotting Incumbents after GLUE Experiments")

    parser.add_argument(
        "--root_dir", type=Path, help="Location of the root directory", default=Path("./")
    )
    parser.add_argument(
        "--benchmark_spec", "-benches",
        nargs="+",
        type=str,
        help="Specification of the benchmark to plot. "
        " (e.g., spec: `benchmark=pd1-cifar100-wide_resnet-2048`, "
        " spec: `benchmark=pd1-cifar100-wide_resnet-2048.objective=valid_error_rate.fidelity=epochs`, " # noqa: E501
        " spec: `benchmark=pd1-imagenet-resnet-512 benchmark=pd1-cifar100-wide_resnet-2048`)"
    )
    parser.add_argument(
        "--optimizer_spec", "-opts",
        type=str,
        nargs="+",
        help="Specification of the optimizer to plot - "
        " (e.g., spec: `optimizer=DEHB`, "
        " spec: `optimizer=DEHB.eta=3`, "
        " spec: `optimizer=DEHB optimizer=SMAC_Hyperband.eta=3`) "
    )
    parser.add_argument(
        "--output_dir",
        type=Path,
        help="Location of the main directory where all studies are stored",
        default=Path.cwd().absolute().parent / "hposuite-output"
    )
    parser.add_argument(
        "--study_dir",
        type=str,
        help="Name of the study directory from where to plot the results",
        default=None
    )
    parser.add_argument(
        "--save_dir",
        type=str,
        help="Directory to save the plots",
        default="plots"
    )
    parser.add_argument(
        "--figsize", "-fs",
        type=int,
        nargs="+",
        default=(20, 10),
        help="Size of the figure to plot",
    )
    parser.add_argument(
        "--logscale", "-ls",
        action="store_true",
        help="Use log scale for the x-axis",
    )
    parser.add_argument(
        "--error_bars", "-eb",
        type=str,
        choices=["std", "sem"],
        default="std",
        help="Type of error bars to plot - "
        "std: Standard deviation, "
        "sem: Standard error of the mean"
    )
    parser.add_argument(
        "--budget_type", "-bt",
        type=str,
        choices=["TrialBudget", "FidelityBudget", None],
        default=None,
        help="Type of budget to plot. "
        "If the study contains a mix of Blackbox and MF opts, "
        "Blackbox opts are only plotted using TrialBudget separately. "
        "MF opts are still plotted using FidelityBudget."
    )
    args = parser.parse_args()

    study_dir = args.output_dir / args.study_dir
    save_dir = study_dir / args.save_dir
    figsize = tuple(map(int, args.figsize))

    agg_data(
        study_dir=study_dir,
        save_dir=save_dir,
        figsize=figsize,
        logscale=args.logscale,
        benchmark_spec=args.benchmark_spec,
        optimizer_spec=args.optimizer_spec,
        error_bars=args.error_bars,
        budget_type=args.budget_type,
    )



from __future__ import annotations

import warnings
from functools import partial
from pathlib import Path

import numpy as np
import pandas as pd
from hpoglue import BenchmarkDescription, Config, Measure, TabularBenchmark
from hpoglue.env import Env
from hpoglue.fidelity import RangeFidelity


def _get_mfh_space(
    function_name: str,
    datadir: str | Path | None = None,
) -> list[Config]:
    """Returns a list of all possible configurations for the MF Hartmann's Tabular Benchmark."""
    table = _get_mfh_table(function_name, datadir)
    config_keys = [k for k in table.columns if "X" in k]
    return TabularBenchmark.get_tabular_config_space(table, config_keys)


def _get_mfh_table(
    function_name: str,
    datadir: str | Path | None = None,
) -> pd.DataFrame:
    return pd.read_parquet(datadir / f"{function_name}.parquet")


def _get_mfh_tabular_benchmark(
    description: BenchmarkDescription,
    *,
    function_name: str,
    datadir: str | Path | None = None,
) -> TabularBenchmark:
    """Creates a TabularBenchmark object for the MF-Hartmann Tabular Benchmark Suite."""
    bench = _get_mfh_table(function_name, datadir)
    config_keys = [k for k in bench.columns if "x" in k]
    return TabularBenchmark(
        desc=description,
        table=bench,
        id_key="config_id",
        config_keys=config_keys,
    )


def mfh_tabular_desc(datadir: str | Path | None = None) -> BenchmarkDescription:
    """Generates benchmark descriptions for the Synthetic MF-Hartmann Tabular Benchmark suite.

    This function iterates over all combinations of function IDs, dimensions, and instances
    defined in `mfh_functions_dict`, `mfh_dims`, and `mfh_instances` respectively. For each
    combination, it creates a `BenchmarkDescription` object with the corresponding configuration
    space, benchmark function, and environment settings.
    """
    for correlation in ("bad", "good", "moderate", "terrible"):
        for dims in (3, 6):
            name = f"mfh{dims}_{correlation}"
            _min = -3.32237 if dims == 3 else -3.86278  # noqa: PLR2004
            try:
                space = _get_mfh_space(name, datadir)
            except FileNotFoundError:
                warnings.warn(  # noqa: B028
                    f"Could not find mfh Tabular Benchmark data for {name}. Skipping. "
                    f"Run `python -m hposuite.benchmarks.create_tabular --benchmark {name} "
                    f"-suite mfh_tabular --task {name}` to create the benchmark data."
                )
                continue
            yield BenchmarkDescription(
                name=f"mfh_tabular-{name}",
                config_space=space,
                load=partial(
                    _get_mfh_tabular_benchmark,
                    function_name=name,
                    datadir=datadir
                ),
                costs={
                    "fid_cost": Measure.cost((0.05, 1), minimize=True),
                },
                fidelities={
                    "z": RangeFidelity.from_tuple((1, 100, 1), supports_continuation=True),
                },
                metrics={
                    "value": Measure.metric((_min, np.inf), minimize=True),
                },
                has_conditionals=False,
                is_tabular=False,
                env=Env.empty(),
                mem_req_mb = 1024,
            )


def mfh_tabular_benchmarks(datadir: str | Path | None = None) :
    """A generator that yields all mfh Tabular benchmarks."""
    if isinstance(datadir, str):
        datadir = Path(datadir).resolve()

    if datadir is None:
        datadir = Path("data", "mfh_tabular").resolve()

    yield from mfh_tabular_desc(datadir=datadir)


from __future__ import annotations

from pathlib import Path

DEFAULT_STUDY_DIR = Path.cwd().absolute().parent / "hposuite-output"
ROOT_DIR = Path(__file__).parent.parent.resolve()
DATA_DIR = ROOT_DIR / "data"


from __future__ import annotations

import copy
from collections.abc import Mapping
from pathlib import Path
from typing import TYPE_CHECKING, Literal

import numpy as np
from ConfigSpace import ConfigurationSpace
from hpoglue import Config, Optimizer, Problem, Query

if TYPE_CHECKING:
    from hpoglue import Result


class RandomSearch(Optimizer):
    """Random Search Optimizer."""

    name = "RandomSearch"

    # NOTE(eddiebergman): Random search doesn't directly use any of this
    # information but we allow it to be used as it's a common baseline.
    support = Problem.Support(
        fidelities=(None,),
        objectives=("single", "many"),
        cost_awareness=(None, "single", "many"),
        tabular=True,
    )

    mem_req_mb = 100

    def __init__(
        self,
        *,
        problem: Problem,
        seed: int,
        working_directory: Path,  # noqa: ARG002
    ):
        """Create a Random Search Optimizer instance for a given problem statement."""
        config_space = problem.config_space
        match config_space:
            case ConfigurationSpace():
                self.config_space = copy.deepcopy(config_space)
                self.config_space.seed(seed)
            case list():
                self.config_space = config_space
            case _:
                raise TypeError("Config space must be a ConfigSpace or a list of Configs")

        self.problem = problem
        self._counter = 0
        self.rng = np.random.default_rng(seed)

    def ask(self) -> Query:
        """Ask the optimizer for a new config to evaluate."""
        self._counter += 1
        # We are dealing with a tabular benchmark
        match self.config_space:
            case ConfigurationSpace():
                config = Config(
                    config_id=str(self._counter),
                    values=dict(self.config_space.sample_configuration()),
                )
            case list():
                index = self.rng.integers(len(self.config_space))
                config = self.config_space[index]
            case _:
                raise TypeError("Config space must be a ConfigSpace or a list of Configs")

        match self.problem.fidelities:
            case None:
                fidelity = None
            case (name, fidelity):
                fidelity = (name, fidelity.max)
            case Mapping():
                fidelity = {
                    name: fidelity.max for name, fidelity in self.problem.fidelities.items()
                }
            case _:
                raise ValueError("Fidelity must be a string or a list of strings")

        return Query(config=config, fidelity=fidelity)

    def tell(self, result: Result) -> None:
        """Tell the optimizer the result of the query."""
        # NOTE(eddiebergman): Random search does nothing with the result


class RandomSearchWithPriors(Optimizer):
    """Random Search Optimizer with Priors over objectives."""

    name = "RandomSearchWithPriors"

    support = Problem.Support(
        objectives=("single", "many"),
        fidelities=(None),
        cost_awareness=(None),
        tabular=False,
        priors=True,
    )

    mem_req_mb = 1024

    def __init__(  # noqa: D107
        self,
        problem: Problem,
        seed: int,
        working_directory: Path,  # noqa: ARG002
        mo_prior_sampling: Literal["random", "equal"] = "random",
    ) -> None:
        self.config_space = problem.config_space
        self.config_space.seed(seed)
        self.problem = problem
        _priors: Mapping[str, Config] = self.problem.priors[1]
        self.priors = {
            obj: _create_normal_prior(
                config_space=self.config_space,
                seed=seed,
                prior_defaults=prior,
                std=0.25,
            )
            for obj, prior in _priors.items()
        }
        self._rng = np.random.default_rng(seed)
        if mo_prior_sampling == "equal":
            assert len(self.problem.get_objectives()) <= self.problem.budget.total, (
                "When using `mo_prior_sampling='equal'` the number of objectives "
                "should be less than or equal to the total budget."
            )
        self.mo_prior_sampling = mo_prior_sampling
        self._optmizer_unique_id = 0
        self._priors_used = {key: 0 for key in self.priors}

    def ask(self) -> Query:
        """Ask the optimizer for a new config to evaluate."""
        self._optmizer_unique_id += 1
        if len(self.priors) > 1:
            match self.mo_prior_sampling:
                case "random":
                    prior = self._rng.choice(list(self.priors.values()))
                case "equal":
                    # raise NotImplementedError
                    min_usage = min(self._priors_used.values())
                    eligible_priors = [
                        key for key, count in self._priors_used.items()
                        if count == min_usage
                    ]
                    selected_prior_key = self._rng.choice(eligible_priors)
                    prior = self.priors[selected_prior_key]
                    self._priors_used[selected_prior_key] += 1
                case _:
                    raise ValueError(
                        "Invalid value for `mo_prior_sampling`. "
                        "Expected one of ['random', 'equal']."
                        f"Got {self.mo_prior_sampling}."
                    )
        else:
            prior = next(iter(self.priors.values()))
        config = Config(
            config_id=str(self._optmizer_unique_id),
            values=dict(prior.sample_configuration()),
        )
        return Query(config=config, fidelity=None)

    def tell(self, result: Result) -> None:
        """Tell the optimizer the result of the query."""
        # NOTE(eddiebergman): Random search does nothing with the result



def _create_normal_prior(
    config_space: ConfigurationSpace,
    seed: int,
    prior_defaults: Config,
    std: float,
) -> ConfigurationSpace:
    from ConfigSpace import (
    BetaFloatHyperparameter,
    BetaIntegerHyperparameter,
    CategoricalHyperparameter,
    Constant,
    NormalFloatHyperparameter,
    NormalIntegerHyperparameter,
    OrdinalHyperparameter,
    UniformFloatHyperparameter,
    UniformIntegerHyperparameter,
)
    prior_space = ConfigurationSpace(seed=seed)
    for hp in list(config_space.values()):
        _default = prior_defaults.values[hp.name]
        match hp:
            case UniformFloatHyperparameter() | BetaFloatHyperparameter():
                prior_space.add(
                    NormalFloatHyperparameter(
                        name=hp.name,
                        mu=_default,
                        sigma=std,
                        lower=hp.lower,
                        upper=hp.upper,
                        default_value=_default,
                        log=hp.log,
                    )
                )
            case UniformIntegerHyperparameter() | BetaIntegerHyperparameter():
                prior_space.add(
                    NormalIntegerHyperparameter(
                        name=hp.name,
                        mu=_default,
                        sigma=std,
                        lower=hp.lower,
                        upper=hp.upper,
                        default_value=_default,
                        log=hp.log,
                    )
                )
            case CategoricalHyperparameter() | OrdinalHyperparameter():
                hp.default_value = _default
                prior_space.add(hp)
            case Constant():
                prior_space.add(hp)
            case NormalFloatHyperparameter() | NormalIntegerHyperparameter():
                hp.mu = _default
                hp.sigma = std
                hp.default_value = _default
                prior_space.add(hp)
            case _:
                raise ValueError(
                    f"Unsupported hyperparameter type: {type(hp).__name__}"
                )
    return prior_space


from __future__ import annotations

from collections.abc import Hashable, Mapping
from pathlib import Path
from typing import TYPE_CHECKING, Any, NoReturn

from ConfigSpace import ConfigurationSpace
from hpoglue import Config, Optimizer, Problem, Query
from hpoglue.budget import CostBudget, TrialBudget
from hpoglue.env import Env
from smac import BlackBoxFacade, HyperbandFacade, MultiFidelityFacade, Scenario
from smac.runhistory import StatusType, TrialValue

if TYPE_CHECKING:
    from hpoglue import Result
    from hpoglue.fidelity import Fidelity
    from smac.facade import AbstractFacade
    from smac.runhistory import TrialInfo


def _dummy_target_function(*args: Any, budget: int | float, seed: int) -> NoReturn:  # noqa: ARG001
    raise RuntimeError("This should never be called!")


class SMAC_Optimizer(Optimizer):
    """Default SMAC Optimizer."""

    env = Env(
        name="SMAC-2.1",
        python_version="3.10",
        requirements=("smac==2.1",)
    )


    def __init__(
        self,
        *,
        problem: Problem,
        seed: int,
        working_directory: Path,
        optimizer: AbstractFacade,
        fidelity: Fidelity | None,
    ):
        """Create a SMAC Optimizer instance for a given problem statement.

        Args:
            problem: Problem statement.

            seed: Random seed for the optimizer.

            working_directory: Working directory to store SMAC run.

            config_space: Configuration space to optimize over.

            optimizer: SMAC optimizer instance.

            fidelity: Fidelity space to optimize over, if any.
        """
        self.problem = problem
        self.working_directory = working_directory
        self.optimizer = optimizer
        self.config_space = problem.config_space
        self._trial_lookup: dict[Hashable, TrialInfo] = {}
        self._fidelity = fidelity
        self._seed = seed

    def ask(self) -> Query:
        """Ask SMAC for a new config to evaluate."""
        smac_info = self.optimizer.ask()
        assert smac_info.instance is None, "We don't do instance benchmarks!"

        config = smac_info.config
        raw_config = dict(config)
        config_id = str(self.optimizer.intensifier.runhistory.config_ids[config])

        fidelity = smac_info.budget

        match fidelity:
            case None:
                fidelity = None
            case float() | int():
                assert self._fidelity is not None
                assert isinstance(self.problem.fidelities, tuple)
                fidelity_value = int(fidelity) if self._fidelity.kind is int else fidelity
                fidelity_name = self.problem.fidelities[0]
                fidelity = (fidelity_name, fidelity_value)
            case _:
                raise NotImplementedError("Unexpected return type for SMAC budget!")

        return Query(
            config=Config(config_id=config_id, values=raw_config),
            fidelity=fidelity,
            optimizer_info=smac_info,
        )

    def tell(self, result: Result) -> None:
        """Tell SMAC the result of the query."""
        match self.problem.objectives:
            case Mapping():
                cost = [
                    obj.as_minimize(result.values[key])
                    for key, obj in self.problem.objectives.items()
                ]
            case (key, obj):
                cost = obj.as_minimize(result.values[key])
            case _:
                raise TypeError("Objective must be a string or a list of strings")

        self.optimizer.tell(
            result.query.optimizer_info,  # type: ignore
            TrialValue(
                cost=cost,
                time=0.0,
                starttime=0.0,
                endtime=0.0,
                status=StatusType.SUCCESS,
                additional_info={},
            ),
            save=True,
        )


class SMAC_BO(SMAC_Optimizer):
    """SMAC Bayesian Optimization."""

    name = "SMAC_BO"

    support = Problem.Support(
        fidelities=(None,),
        objectives=("single", "many"),
        cost_awareness=(None,),
        tabular=False,
    )

    mem_req_mb = 1024

    def __init__(
        self,
        *,
        problem: Problem,
        seed: int,
        working_directory: Path,
        xi: float = 0.0,
    ):
        """Create a SMAC BO Optimizer instance for a given problem statement.

        Args:
            problem: Problem statement.

            seed: Random seed for the optimizer.

            working_directory: Working directory to store SMAC run.

            xi: Exploration-exploitation trade-off parameter. Defaults to 0.0.
        """
        config_space = problem.config_space
        match config_space:
            case ConfigurationSpace():
                pass
            case list():
                raise ValueError("SMAC does not support tabular benchmarks!")
            case _:
                raise TypeError("Config space must be a list or a ConfigurationSpace!")

        match problem.fidelities:
            case None:
                pass
            case tuple() | Mapping():
                raise ValueError("SMAC BO does not support multi-fidelity benchmarks!")
            case _:
                raise TypeError("Fidelity must be a string or a list of strings!")

        match problem.objectives:
            case Mapping():
                metric_names = list(problem.objectives.keys())
            case (metric_name, _):
                metric_names = metric_name
            case _:
                raise TypeError("Objective must be a tuple of (name, metric) or a mapping")

        working_directory.mkdir(parents=True, exist_ok=True)

        match problem.budget:
            case TrialBudget():
                budget = problem.budget.total
            case CostBudget():
                raise ValueError("SMAC BO does not support cost-aware benchmarks!")
            case _:
                raise TypeError("Budget must be a TrialBudget or a CostBudget!")

        scenario = Scenario(
            configspace=config_space,
            deterministic=True,
            objectives=metric_names,
            n_trials=budget,
            seed=seed,
            output_directory=working_directory / "smac-output",
            min_budget=None,
            max_budget=None,
        )
        super().__init__(
            problem=problem,
            seed=seed,
            working_directory=working_directory,
            fidelity=None,
            optimizer=BlackBoxFacade(
                scenario=scenario,
                logging_level=False,
                target_function=_dummy_target_function,
                intensifier=BlackBoxFacade.get_intensifier(scenario),
                acquisition_function=BlackBoxFacade.get_acquisition_function(scenario, xi=xi),
                overwrite=True,
            ),
        )


class SMAC_Hyperband(SMAC_Optimizer):
    """SMAC Hyperband."""

    name = "SMAC_Hyperband"
    support = Problem.Support(
        fidelities=("single",),
        objectives=("single", "many"),
        cost_awareness=(None,),
        tabular=False,
    )
    mem_req_mb = 1024

    def __init__(  # noqa: PLR0912
        self,
        *,
        problem: Problem,
        seed: int,
        working_directory: Path,
        eta: int = 3,
    ):
        """Create a SMAC Hyperband Optimizer instance for a given problem statement.

        Args:
            problem: Problem statement.

            seed: Random seed for the optimizer.

            working_directory: Working directory to store SMAC run.

            eta: Hyperband eta parameter. Defaults to 3.
        """
        config_space = problem.config_space
        match config_space:
            case ConfigurationSpace():
                pass
            case list():
                raise ValueError("SMAC does not support tabular benchmarks!")
            case _:
                raise TypeError("Config space must be a list or a ConfigurationSpace!")

        min_fidelity: float | int
        max_fidelity: float | int
        match problem.fidelities:
            case None:
                raise ValueError("SMAC Hyperband requires a fidelity space!")
            case Mapping():
                raise ValueError("SMAC Hyperband does not support many-fidelity!")
            case (_, fidelity):
                _fid = fidelity
                min_fidelity = fidelity.min
                max_fidelity = fidelity.max
            case _:
                raise TypeError("Fidelity must be a string or a list of strings!")

        match problem.objectives:
            case Mapping():
                metric_names = list(problem.objectives.keys())
            case (metric_name, _):
                metric_names = metric_name
            case _:
                raise TypeError("Objective must be a tuple of (name, metric) or a mapping")

        working_directory.mkdir(parents=True, exist_ok=True)
        match problem.budget:
            case TrialBudget():
                budget = problem.budget.total
            case CostBudget():
                raise ValueError("SMAC Hyperband does not support cost-aware benchmarks!")
            case _:
                raise TypeError("Budget must be a TrialBudget or a CostBudget!")

        scenario = Scenario(
            configspace=config_space,
            deterministic=True,
            objectives=metric_names,
            n_trials=budget,
            seed=seed,
            output_directory=working_directory / "smac-output",
            min_budget=min_fidelity,
            max_budget=max_fidelity,
        )
        super().__init__(
            problem=problem,
            seed=seed,
            working_directory=working_directory,
            fidelity=_fid,
            optimizer=HyperbandFacade(
                scenario=scenario,
                logging_level=False,
                target_function=_dummy_target_function,
                intensifier=HyperbandFacade.get_intensifier(scenario, eta=eta),
                overwrite=True,
            ),
        )


class SMAC_BOHB(SMAC_Optimizer):
    """SMAC BOHB."""

    name = "SMAC_BOHB"
    support = Problem.Support(
        fidelities=("single",),
        objectives=("single", "many"),
        cost_awareness=(None,),
        tabular=False,
    )
    mem_req_mb = 1024

    def __init__(  # noqa: PLR0912
        self,
        *,
        problem: Problem,
        seed: int,
        working_directory: Path,
        eta: int = 3,
    ):
        """Create a SMAC BOHB Optimizer instance for a given problem statement.

        Args:
            problem: Problem statement.

            seed: Random seed for the optimizer.

            working_directory: Working directory to store SMAC run.

            eta: BOHB eta parameter. Defaults to 3.
                According to the paper, changing this doesn't seem to make a difference.
        """
        config_space = problem.config_space
        match config_space:
            case ConfigurationSpace():
                pass
            case list():
                raise ValueError("SMAC does not support tabular benchmarks!")
            case _:
                raise TypeError("Config space must be a list or a ConfigurationSpace!")

        min_fidelity: float | int
        max_fidelity: float | int
        match problem.fidelities:
            case None:
                raise ValueError("SMAC BOHB requires a fidelity space!")
            case Mapping():
                raise ValueError("SMAC BOHB does not support many-fidelity!")
            case (_, fidelity):
                _fid = fidelity
                min_fidelity = fidelity.min
                max_fidelity = fidelity.max
            case _:
                raise TypeError("Fidelity must be a string or a list of strings!")

        match problem.objectives:
            case Mapping():
                metric_names = list(problem.objectives.keys())
            case (metric_name, _):
                metric_names = metric_name
            case _:
                raise TypeError("Objective must be a tuple of (name, metric) or a mapping")

        working_directory.mkdir(parents=True, exist_ok=True)
        match problem.budget:
            case TrialBudget():
                budget = problem.budget.total
            case CostBudget():
                raise ValueError("SMAC BOHB does not support cost-aware benchmarks!")
            case _:
                raise TypeError("Budget must be a TrialBudget or a CostBudget!")

        scenario = Scenario(
            configspace=config_space,
            deterministic=True,
            objectives=metric_names,
            n_trials=budget,
            seed=seed,
            output_directory=working_directory / "smac-output",
            min_budget=min_fidelity,
            max_budget=max_fidelity,
        )
        super().__init__(
            problem=problem,
            seed=seed,
            working_directory=working_directory,
            fidelity=_fid,
            optimizer=MultiFidelityFacade(
                scenario=scenario,
                logging_level=False,
                target_function=_dummy_target_function,
                intensifier=MultiFidelityFacade.get_intensifier(scenario, eta=eta),
                overwrite=True,
            ),
        )



from __future__ import annotations

# ruff: noqa: PD901
from typing import TYPE_CHECKING

import numpy as np
import pandas as pd

if TYPE_CHECKING:
    import matplotlib.pyplot as plt

SEED_COL = "run.seed"
PROBLEM_COL = "problem.name"
OPTIMIZER_COL = "optimizer.name"
HP_COL = "run.opt.hp_str"
SINGLE_OBJ_COL = "result.objective.1.value"
SECOND_OBJ_COL = "result.objective.2.value"
BUDGET_USED_COL = "result.budget_used_total"
BUDGET_TOTAL_COL = "problem.budget.total"


def _plot_single_objective_incumbent_trace_for_problem_and_optimizer(
    df: pd.DataFrame,
    *,
    ycol: str,
    ax: plt.Axes,
    to_minimize: bool,
    regret_bound: float,
    marker: str,
    hue: str,
    label: str,
) -> None:
    assert len(df) > 0
    assert PROBLEM_COL in df.columns

    problem_names = df[PROBLEM_COL]
    assert (problem_names[0] == df[PROBLEM_COL]).all()
    optimizer_names = df[OPTIMIZER_COL]
    assert (optimizer_names[0] == df[OPTIMIZER_COL]).all()
    optimizer_hps = df[HP_COL]
    assert (optimizer_hps[0] == df[HP_COL]).all()
    assert ycol in df.columns

    match to_minimize, regret_bound:
        case True, None:
            assert df[ycol].is_monotonic_decreasing
            df["y"] = df[ycol]
        case True, _:
            assert df[ycol].is_monotonic_decreasing
            df["y"] = df[ycol] - regret_bound
        case False, None:
            assert df[ycol].is_monotonic_increasing
            df["y"] = df[ycol]
        case False, _:
            assert df[ycol].is_monotonic_increasing
            df["y"] = regret_bound - df[ycol]
        case _:
            raise ValueError("Invalid combination of `to_minimize` and `regret_bound`")

    if SEED_COL in df.columns:
        full_frame = pd.DataFrame()
        for seed, seed_df in df.groupby(SEED_COL):
            full_frame[seed] = seed_df[[BUDGET_USED_COL, "y"]].set_index(BUDGET_USED_COL)

        _data = full_frame.sort_index().ffill().dropna()
        _data = _data.agg(["mean", "std"], axis=1)

        _xs = _data.index.astype(np.float32).to_numpy()
        _means = _data["mean"].to_numpy()  # type: ignore
        _stds = _data["std"].to_numpy()  # type: ignore
    else:
        _xs = df[BUDGET_USED_COL].astype(np.float32).to_numpy()
        _means = df["y"].to_numpy()
        _stds = None

    ax.plot(  # type: ignore
        _xs,
        _means,
        drawstyle="steps-post",
        label=label,
        linestyle="solid",  # type: ignore
        markevery=10,
        marker=marker,
        linewidth=3,
    )

    if _stds is not None:
        ax.fill_between(
            _xs,
            _means - _stds,
            _means + _stds,
            alpha=0.2,
            color=hue,
            edgecolor=hue,
            linewidth=2,
            step="post",
        )
